# test_refs.py -- tests for refs.py
# Copyright (C) 2013 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for dulwich.refs."""

import os
import sys
import tempfile
from io import BytesIO
from typing import ClassVar

from dulwich import errors
from dulwich.file import GitFile
from dulwich.objects import ZERO_SHA
from dulwich.refs import (
    DictRefsContainer,
    InfoRefsContainer,
    SymrefLoop,
    _split_ref_line,
    check_ref_format,
    is_per_worktree_ref,
    parse_remote_ref,
    parse_symref_value,
    read_packed_refs,
    read_packed_refs_with_peeled,
    shorten_ref_name,
    split_peeled_refs,
    strip_peeled_refs,
    write_packed_refs,
)
from dulwich.repo import Repo
from dulwich.tests.utils import open_repo, tear_down_repo
from dulwich.worktree import add_worktree

from . import SkipTest, TestCase


class CheckRefFormatTests(TestCase):
    """Tests for the check_ref_format function.

    These are the same tests as in the git test suite.
    """

    def test_valid(self) -> None:
        self.assertTrue(check_ref_format(b"heads/foo"))
        self.assertTrue(check_ref_format(b"foo/bar/baz"))
        self.assertTrue(check_ref_format(b"refs///heads/foo"))
        self.assertTrue(check_ref_format(b"foo./bar"))
        self.assertTrue(check_ref_format(b"heads/foo@bar"))
        self.assertTrue(check_ref_format(b"heads/fix.lock.error"))

    def test_invalid(self) -> None:
        self.assertFalse(check_ref_format(b"foo"))
        self.assertFalse(check_ref_format(b"heads/foo/"))
        self.assertFalse(check_ref_format(b"./foo"))
        self.assertFalse(check_ref_format(b".refs/foo"))
        self.assertFalse(check_ref_format(b"heads/foo..bar"))
        self.assertFalse(check_ref_format(b"heads/foo?bar"))
        self.assertFalse(check_ref_format(b"heads/foo.lock"))
        self.assertFalse(check_ref_format(b"heads/v@{ation"))
        self.assertFalse(check_ref_format(b"heads/foo\bar"))


ONES = b"1" * 40
TWOS = b"2" * 40
THREES = b"3" * 40
FOURS = b"4" * 40


class PackedRefsFileTests(TestCase):
    def test_split_ref_line_errors(self) -> None:
        self.assertRaises(errors.PackedRefsException, _split_ref_line, b"singlefield")
        self.assertRaises(errors.PackedRefsException, _split_ref_line, b"badsha name")
        self.assertRaises(
            errors.PackedRefsException,
            _split_ref_line,
            ONES + b" bad/../refname",
        )

    def test_read_without_peeled(self) -> None:
        f = BytesIO(b"\n".join([b"# comment", ONES + b" ref/1", TWOS + b" ref/2"]))
        self.assertEqual(
            [(ONES, b"ref/1"), (TWOS, b"ref/2")], list(read_packed_refs(f))
        )

    def test_read_without_peeled_errors(self) -> None:
        f = BytesIO(b"\n".join([ONES + b" ref/1", b"^" + TWOS]))
        self.assertRaises(errors.PackedRefsException, list, read_packed_refs(f))

    def test_read_with_peeled(self) -> None:
        f = BytesIO(
            b"\n".join(
                [
                    ONES + b" ref/1",
                    TWOS + b" ref/2",
                    b"^" + THREES,
                    FOURS + b" ref/4",
                ]
            )
        )
        self.assertEqual(
            [
                (ONES, b"ref/1", None),
                (TWOS, b"ref/2", THREES),
                (FOURS, b"ref/4", None),
            ],
            list(read_packed_refs_with_peeled(f)),
        )

    def test_read_with_peeled_errors(self) -> None:
        f = BytesIO(b"\n".join([b"^" + TWOS, ONES + b" ref/1"]))
        self.assertRaises(errors.PackedRefsException, list, read_packed_refs(f))

        f = BytesIO(b"\n".join([ONES + b" ref/1", b"^" + TWOS, b"^" + THREES]))
        self.assertRaises(errors.PackedRefsException, list, read_packed_refs(f))

    def test_write_with_peeled(self) -> None:
        f = BytesIO()
        write_packed_refs(f, {b"ref/1": ONES, b"ref/2": TWOS}, {b"ref/1": THREES})
        self.assertEqual(
            b"\n".join(
                [
                    b"# pack-refs with: peeled",
                    ONES + b" ref/1",
                    b"^" + THREES,
                    TWOS + b" ref/2",
                ]
            )
            + b"\n",
            f.getvalue(),
        )

    def test_write_without_peeled(self) -> None:
        f = BytesIO()
        write_packed_refs(f, {b"ref/1": ONES, b"ref/2": TWOS})
        self.assertEqual(
            b"\n".join([ONES + b" ref/1", TWOS + b" ref/2"]) + b"\n",
            f.getvalue(),
        )


# Dict of refs that we expect all RefsContainerTests subclasses to define.
_TEST_REFS = {
    b"HEAD": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/heads/40-char-ref-aaaaaaaaaaaaaaaaaa": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/heads/master": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
    b"refs/tags/refs-0.1": b"df6800012397fb85c56e7418dd4eb9405dee075c",
    b"refs/tags/refs-0.2": b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8",
    b"refs/heads/loop": b"ref: refs/heads/loop",
}


class RefsContainerTests:
    def test_keys(self) -> None:
        actual_keys = set(self._refs.keys())
        self.assertEqual(set(self._refs.allkeys()), actual_keys)
        self.assertEqual(set(_TEST_REFS.keys()), actual_keys)

        actual_keys = self._refs.keys(b"refs/heads")
        actual_keys.discard(b"loop")
        self.assertEqual(
            [b"40-char-ref-aaaaaaaaaaaaaaaaaa", b"master", b"packed"],
            sorted(actual_keys),
        )
        self.assertEqual(
            [b"refs-0.1", b"refs-0.2"], sorted(self._refs.keys(b"refs/tags"))
        )

    def test_iter(self) -> None:
        actual_keys = set(self._refs.keys())
        self.assertEqual(set(self._refs), actual_keys)
        self.assertEqual(set(_TEST_REFS.keys()), actual_keys)

    def test_as_dict(self) -> None:
        # refs/heads/loop does not show up even if it exists
        expected_refs = dict(_TEST_REFS)
        del expected_refs[b"refs/heads/loop"]
        self.assertEqual(expected_refs, self._refs.as_dict())

    def test_get_symrefs(self) -> None:
        self._refs.set_symbolic_ref(b"refs/heads/src", b"refs/heads/dst")
        symrefs = self._refs.get_symrefs()
        if b"HEAD" in symrefs:
            symrefs.pop(b"HEAD")
        self.assertEqual(
            {
                b"refs/heads/src": b"refs/heads/dst",
                b"refs/heads/loop": b"refs/heads/loop",
            },
            symrefs,
        )

    def test_setitem(self) -> None:
        self._refs[b"refs/some/ref"] = b"42d06bd4b77fed026b154d16493e5deab78f02ec"
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/some/ref"],
        )

        # should accept symref
        self._refs[b"refs/heads/symbolic"] = b"ref: refs/heads/master"
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/symbolic"],
        )

        # should not accept bad ref names
        self.assertRaises(
            errors.RefFormatError,
            self._refs.__setitem__,
            b"notrefs/foo",
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
        )

        # should not accept short sha
        self.assertRaises(
            ValueError,
            self._refs.__setitem__,
            b"refs/some/ref",
            b"42d06bd",
        )

    def test_set_if_equals(self) -> None:
        nines = b"9" * 40
        self.assertFalse(self._refs.set_if_equals(b"HEAD", b"c0ffee", nines))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec", self._refs[b"HEAD"]
        )

        self.assertTrue(
            self._refs.set_if_equals(
                b"HEAD", b"42d06bd4b77fed026b154d16493e5deab78f02ec", nines
            )
        )
        self.assertEqual(nines, self._refs[b"HEAD"])

        # Setting the ref again is a no-op, but will return True.
        self.assertTrue(self._refs.set_if_equals(b"HEAD", nines, nines))
        self.assertEqual(nines, self._refs[b"HEAD"])

        self.assertTrue(self._refs.set_if_equals(b"refs/heads/master", None, nines))
        self.assertEqual(nines, self._refs[b"refs/heads/master"])

        self.assertTrue(
            self._refs.set_if_equals(b"refs/heads/nonexistent", ZERO_SHA, nines)
        )
        self.assertEqual(nines, self._refs[b"refs/heads/nonexistent"])

    def test_add_if_new(self) -> None:
        nines = b"9" * 40
        self.assertFalse(self._refs.add_if_new(b"refs/heads/master", nines))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/master"],
        )

        self.assertTrue(self._refs.add_if_new(b"refs/some/ref", nines))
        self.assertEqual(nines, self._refs[b"refs/some/ref"])

    def test_set_symbolic_ref(self) -> None:
        self._refs.set_symbolic_ref(b"refs/heads/symbolic", b"refs/heads/master")
        self.assertEqual(
            b"ref: refs/heads/master",
            self._refs.read_loose_ref(b"refs/heads/symbolic"),
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/symbolic"],
        )

    def test_set_symbolic_ref_overwrite(self) -> None:
        nines = b"9" * 40
        self.assertNotIn(b"refs/heads/symbolic", self._refs)
        self._refs[b"refs/heads/symbolic"] = nines
        self.assertEqual(nines, self._refs.read_loose_ref(b"refs/heads/symbolic"))
        self._refs.set_symbolic_ref(b"refs/heads/symbolic", b"refs/heads/master")
        self.assertEqual(
            b"ref: refs/heads/master",
            self._refs.read_loose_ref(b"refs/heads/symbolic"),
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/symbolic"],
        )

    def test_check_refname(self) -> None:
        self._refs._check_refname(b"HEAD")
        self._refs._check_refname(b"refs/stash")
        self._refs._check_refname(b"refs/heads/foo")

        self.assertRaises(errors.RefFormatError, self._refs._check_refname, b"refs")
        self.assertRaises(
            errors.RefFormatError, self._refs._check_refname, b"notrefs/foo"
        )

    def test_contains(self) -> None:
        self.assertIn(b"refs/heads/master", self._refs)
        self.assertNotIn(b"refs/heads/bar", self._refs)

    def test_delitem(self) -> None:
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/master"],
        )
        del self._refs[b"refs/heads/master"]
        self.assertRaises(KeyError, lambda: self._refs[b"refs/heads/master"])

    def test_remove_if_equals(self) -> None:
        self.assertFalse(self._refs.remove_if_equals(b"HEAD", b"c0ffee"))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec", self._refs[b"HEAD"]
        )
        self.assertTrue(
            self._refs.remove_if_equals(
                b"refs/tags/refs-0.2",
                b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8",
            )
        )
        self.assertTrue(self._refs.remove_if_equals(b"refs/tags/refs-0.2", ZERO_SHA))
        self.assertNotIn(b"refs/tags/refs-0.2", self._refs)

    def test_import_refs_name(self) -> None:
        self._refs[b"refs/remotes/origin/other"] = (
            b"48d01bd4b77fed026b154d16493e5deab78f02ec"
        )
        self._refs.import_refs(
            b"refs/remotes/origin",
            {b"master": b"42d06bd4b77fed026b154d16493e5deab78f02ec"},
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/remotes/origin/master"],
        )
        self.assertEqual(
            b"48d01bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/remotes/origin/other"],
        )

    def test_import_refs_name_prune(self) -> None:
        self._refs[b"refs/remotes/origin/other"] = (
            b"48d01bd4b77fed026b154d16493e5deab78f02ec"
        )
        self._refs.import_refs(
            b"refs/remotes/origin",
            {b"master": b"42d06bd4b77fed026b154d16493e5deab78f02ec"},
            prune=True,
        )
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/remotes/origin/master"],
        )
        self.assertNotIn(b"refs/remotes/origin/other", self._refs)


class DictRefsContainerTests(RefsContainerTests, TestCase):
    def setUp(self) -> None:
        TestCase.setUp(self)
        self._refs = DictRefsContainer(dict(_TEST_REFS))

    def test_invalid_refname(self) -> None:
        # FIXME: Move this test into RefsContainerTests, but requires
        # some way of injecting invalid refs.
        self._refs._refs[b"refs/stash"] = b"00" * 20
        expected_refs = dict(_TEST_REFS)
        del expected_refs[b"refs/heads/loop"]
        expected_refs[b"refs/stash"] = b"00" * 20
        self.assertEqual(expected_refs, self._refs.as_dict())

    def test_set_if_equals_with_symbolic_ref(self) -> None:
        # Test that set_if_equals only updates the requested ref,
        # not all refs in a symbolic reference chain

        # The bug in the original implementation was that when follow()
        # was called on a ref, it would return all refs in the chain,
        # and set_if_equals would update ALL of them instead of just the
        # requested ref.

        # Set up refs
        master_sha = b"1" * 40
        feature_sha = b"2" * 40
        new_sha = b"3" * 40

        self._refs[b"refs/heads/master"] = master_sha
        self._refs[b"refs/heads/feature"] = feature_sha
        # Create a second symbolic ref pointing to feature
        self._refs.set_symbolic_ref(b"refs/heads/other", b"refs/heads/feature")

        # Update refs/heads/other through set_if_equals
        # With the bug, this would update BOTH refs/heads/other AND refs/heads/feature
        # Without the bug, only refs/heads/other should be updated
        # Note: old_ref needs to be the actual stored value (the symref)
        self.assertTrue(
            self._refs.set_if_equals(
                b"refs/heads/other", b"ref: refs/heads/feature", new_sha
            )
        )

        # refs/heads/other should now directly point to new_sha
        self.assertEqual(self._refs.read_ref(b"refs/heads/other"), new_sha)

        # refs/heads/feature should remain unchanged
        # With the bug, refs/heads/feature would also be incorrectly updated to new_sha
        self.assertEqual(self._refs[b"refs/heads/feature"], feature_sha)
        self.assertEqual(self._refs[b"refs/heads/master"], master_sha)


class DiskRefsContainerTests(RefsContainerTests, TestCase):
    def setUp(self) -> None:
        TestCase.setUp(self)
        self._repo = open_repo("refs.git")
        self.addCleanup(tear_down_repo, self._repo)
        self._refs = self._repo.refs

    def test_get_packed_refs(self) -> None:
        self.assertEqual(
            {
                b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
                b"refs/tags/refs-0.1": b"df6800012397fb85c56e7418dd4eb9405dee075c",
            },
            self._refs.get_packed_refs(),
        )

    def test_get_peeled_not_packed(self) -> None:
        # not packed
        self.assertEqual(None, self._refs.get_peeled(b"refs/tags/refs-0.2"))
        self.assertEqual(
            b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8",
            self._refs[b"refs/tags/refs-0.2"],
        )

        # packed, known not peelable
        self.assertEqual(
            self._refs[b"refs/heads/packed"],
            self._refs.get_peeled(b"refs/heads/packed"),
        )

        # packed, peeled
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs.get_peeled(b"refs/tags/refs-0.1"),
        )

    def test_setitem(self) -> None:
        RefsContainerTests.test_setitem(self)
        path = os.path.join(self._refs.path, b"refs", b"some", b"ref")
        with open(path, "rb") as f:
            self.assertEqual(b"42d06bd4b77fed026b154d16493e5deab78f02ec", f.read()[:40])

        self.assertRaises(
            OSError,
            self._refs.__setitem__,
            b"refs/some/ref/sub",
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
        )

    def test_delete_refs_container(self) -> None:
        # We shouldn't delete the refs directory
        self._refs[b"refs/heads/blah"] = b"42d06bd4b77fed026b154d16493e5deab78f02ec"
        for ref in self._refs.allkeys():
            del self._refs[ref]
        self.assertTrue(os.path.exists(os.path.join(self._refs.path, b"refs")))

    def test_setitem_packed(self) -> None:
        with open(os.path.join(self._refs.path, b"packed-refs"), "w") as f:
            f.write("# pack-refs with: peeled fully-peeled sorted \n")
            f.write("42d06bd4b77fed026b154d16493e5deab78f02ec refs/heads/packed\n")

        # It's allowed to set a new ref on a packed ref, the new ref will be
        # placed outside on refs/
        self._refs[b"refs/heads/packed"] = b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8"
        packed_ref_path = os.path.join(self._refs.path, b"refs", b"heads", b"packed")
        with open(packed_ref_path, "rb") as f:
            self.assertEqual(b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8", f.read()[:40])

        self.assertRaises(
            OSError,
            self._refs.__setitem__,
            b"refs/heads/packed/sub",
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
        )

        # this shouldn't overwrite the packed refs
        self.assertEqual(
            {b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec"},
            self._refs.get_packed_refs(),
        )

    def test_add_packed_refs(self) -> None:
        # first, create a non-packed ref
        self._refs[b"refs/heads/packed"] = b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8"

        packed_ref_path = os.path.join(self._refs.path, b"refs", b"heads", b"packed")
        self.assertTrue(os.path.exists(packed_ref_path))

        # now overwrite that with a packed ref
        packed_refs_file_path = os.path.join(self._refs.path, b"packed-refs")
        self._refs.add_packed_refs(
            {
                b"refs/heads/packed": b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            }
        )

        # that should kill the file
        self.assertFalse(os.path.exists(packed_ref_path))

        # now delete the packed ref
        self._refs.add_packed_refs(
            {
                b"refs/heads/packed": None,
            }
        )

        # and it's gone!
        self.assertFalse(os.path.exists(packed_ref_path))

        self.assertRaises(
            KeyError,
            self._refs.__getitem__,
            b"refs/heads/packed",
        )

        # just in case, make sure we can't pack HEAD
        self.assertRaises(
            ValueError,
            self._refs.add_packed_refs,
            {b"HEAD": "02ac81614bcdbd585a37b4b0edf8cb8a"},
        )

        # delete all packed refs
        self._refs.add_packed_refs({ref: None for ref in self._refs.get_packed_refs()})

        self.assertEqual({}, self._refs.get_packed_refs())

        # remove the packed ref file, and check that adding nothing doesn't affect that
        os.remove(packed_refs_file_path)

        # adding nothing doesn't make it reappear
        self._refs.add_packed_refs({})

        self.assertFalse(os.path.exists(packed_refs_file_path))

    def test_setitem_symbolic(self) -> None:
        ones = b"1" * 40
        self._refs[b"HEAD"] = ones
        self.assertEqual(ones, self._refs[b"HEAD"])

        # ensure HEAD was not modified
        f = open(os.path.join(self._refs.path, b"HEAD"), "rb")
        v = next(iter(f)).rstrip(b"\n\r")
        f.close()
        self.assertEqual(b"ref: refs/heads/master", v)

        # ensure the symbolic link was written through
        f = open(os.path.join(self._refs.path, b"refs", b"heads", b"master"), "rb")
        self.assertEqual(ones, f.read()[:40])
        f.close()

    def test_set_if_equals(self) -> None:
        RefsContainerTests.test_set_if_equals(self)

        # ensure symref was followed
        self.assertEqual(b"9" * 40, self._refs[b"refs/heads/master"])

        # ensure lockfile was deleted
        self.assertFalse(
            os.path.exists(
                os.path.join(self._refs.path, b"refs", b"heads", b"master.lock")
            )
        )
        self.assertFalse(os.path.exists(os.path.join(self._refs.path, b"HEAD.lock")))

    def test_add_if_new_packed(self) -> None:
        # don't overwrite packed ref
        self.assertFalse(self._refs.add_if_new(b"refs/tags/refs-0.1", b"9" * 40))
        self.assertEqual(
            b"df6800012397fb85c56e7418dd4eb9405dee075c",
            self._refs[b"refs/tags/refs-0.1"],
        )

    def test_add_if_new_symbolic(self) -> None:
        # Use an empty repo instead of the default.
        repo_dir = os.path.join(tempfile.mkdtemp(), "test")
        os.makedirs(repo_dir)
        repo = Repo.init(repo_dir)
        self.addCleanup(tear_down_repo, repo)
        refs = repo.refs

        nines = b"9" * 40
        self.assertEqual(b"ref: refs/heads/master", refs.read_ref(b"HEAD"))
        self.assertNotIn(b"refs/heads/master", refs)
        self.assertTrue(refs.add_if_new(b"HEAD", nines))
        self.assertEqual(b"ref: refs/heads/master", refs.read_ref(b"HEAD"))
        self.assertEqual(nines, refs[b"HEAD"])
        self.assertEqual(nines, refs[b"refs/heads/master"])
        self.assertFalse(refs.add_if_new(b"HEAD", b"1" * 40))
        self.assertEqual(nines, refs[b"HEAD"])
        self.assertEqual(nines, refs[b"refs/heads/master"])

    def test_follow(self) -> None:
        self.assertEqual(
            (
                [b"HEAD", b"refs/heads/master"],
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            ),
            self._refs.follow(b"HEAD"),
        )
        self.assertEqual(
            (
                [b"refs/heads/master"],
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            ),
            self._refs.follow(b"refs/heads/master"),
        )
        self.assertRaises(SymrefLoop, self._refs.follow, b"refs/heads/loop")

    def test_set_overwrite_loop(self) -> None:
        self.assertRaises(SymrefLoop, self._refs.follow, b"refs/heads/loop")
        self._refs[b"refs/heads/loop"] = b"42d06bd4b77fed026b154d16493e5deab78f02ec"
        self.assertEqual(
            ([b"refs/heads/loop"], b"42d06bd4b77fed026b154d16493e5deab78f02ec"),
            self._refs.follow(b"refs/heads/loop"),
        )

    def test_delitem(self) -> None:
        RefsContainerTests.test_delitem(self)
        ref_file = os.path.join(self._refs.path, b"refs", b"heads", b"master")
        self.assertFalse(os.path.exists(ref_file))
        self.assertNotIn(b"refs/heads/master", self._refs.get_packed_refs())

    def test_delitem_symbolic(self) -> None:
        self.assertEqual(b"ref: refs/heads/master", self._refs.read_loose_ref(b"HEAD"))
        del self._refs[b"HEAD"]
        self.assertRaises(KeyError, lambda: self._refs[b"HEAD"])
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs[b"refs/heads/master"],
        )
        self.assertFalse(os.path.exists(os.path.join(self._refs.path, b"HEAD")))

    def test_remove_if_equals_symref(self) -> None:
        # HEAD is a symref, so shouldn't equal its dereferenced value
        self.assertFalse(
            self._refs.remove_if_equals(
                b"HEAD", b"42d06bd4b77fed026b154d16493e5deab78f02ec"
            )
        )
        self.assertTrue(
            self._refs.remove_if_equals(
                b"refs/heads/master",
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            )
        )
        self.assertRaises(KeyError, lambda: self._refs[b"refs/heads/master"])

        # HEAD is now a broken symref
        self.assertRaises(KeyError, lambda: self._refs[b"HEAD"])
        self.assertEqual(b"ref: refs/heads/master", self._refs.read_loose_ref(b"HEAD"))

        self.assertFalse(
            os.path.exists(
                os.path.join(self._refs.path, b"refs", b"heads", b"master.lock")
            )
        )
        self.assertFalse(os.path.exists(os.path.join(self._refs.path, b"HEAD.lock")))

    def test_remove_packed_without_peeled(self) -> None:
        refs_file = os.path.join(self._repo.path, "packed-refs")
        f = GitFile(refs_file)
        refs_data = f.read()
        f.close()
        f = GitFile(refs_file, "wb")
        f.write(
            b"\n".join(
                line
                for line in refs_data.split(b"\n")
                if not line or line[0] not in b"#^"
            )
        )
        f.close()
        self._repo = Repo(self._repo.path)
        refs = self._repo.refs
        self.assertTrue(
            refs.remove_if_equals(
                b"refs/heads/packed",
                b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            )
        )

    def test_remove_if_equals_packed(self) -> None:
        # test removing ref that is only packed
        self.assertEqual(
            b"df6800012397fb85c56e7418dd4eb9405dee075c",
            self._refs[b"refs/tags/refs-0.1"],
        )
        self.assertTrue(
            self._refs.remove_if_equals(
                b"refs/tags/refs-0.1",
                b"df6800012397fb85c56e7418dd4eb9405dee075c",
            )
        )
        self.assertRaises(KeyError, lambda: self._refs[b"refs/tags/refs-0.1"])

    def test_remove_parent(self) -> None:
        self._refs[b"refs/heads/foo/bar"] = b"df6800012397fb85c56e7418dd4eb9405dee075c"
        del self._refs[b"refs/heads/foo/bar"]
        ref_file = os.path.join(
            self._refs.path,
            b"refs",
            b"heads",
            b"foo",
            b"bar",
        )
        self.assertFalse(os.path.exists(ref_file))
        ref_file = os.path.join(self._refs.path, b"refs", b"heads", b"foo")
        self.assertFalse(os.path.exists(ref_file))
        ref_file = os.path.join(self._refs.path, b"refs", b"heads")
        self.assertTrue(os.path.exists(ref_file))
        self._refs[b"refs/heads/foo"] = b"df6800012397fb85c56e7418dd4eb9405dee075c"

    def test_read_ref(self) -> None:
        self.assertEqual(b"ref: refs/heads/master", self._refs.read_ref(b"HEAD"))
        self.assertEqual(
            b"42d06bd4b77fed026b154d16493e5deab78f02ec",
            self._refs.read_ref(b"refs/heads/packed"),
        )
        self.assertEqual(None, self._refs.read_ref(b"nonexistent"))

    def test_read_loose_ref(self) -> None:
        self._refs[b"refs/heads/foo"] = b"df6800012397fb85c56e7418dd4eb9405dee075c"

        self.assertEqual(None, self._refs.read_ref(b"refs/heads/foo/bar"))

    def test_non_ascii(self) -> None:
        try:
            encoded_ref = os.fsencode("refs/tags/schön")
        except UnicodeEncodeError as exc:
            raise SkipTest(
                "filesystem encoding doesn't support special character"
            ) from exc
        p = os.path.join(os.fsencode(self._repo.path), encoded_ref)
        with open(p, "w") as f:
            f.write("00" * 20)

        expected_refs = dict(_TEST_REFS)
        expected_refs[encoded_ref] = b"00" * 20
        del expected_refs[b"refs/heads/loop"]

        self.assertEqual(expected_refs, self._repo.get_refs())

    def test_cyrillic(self) -> None:
        if sys.platform in ("darwin", "win32"):
            raise SkipTest("filesystem encoding doesn't support arbitrary bytes")
        # reported in https://github.com/dulwich/dulwich/issues/608
        name = b"\xcd\xee\xe2\xe0\xff\xe2\xe5\xf2\xea\xe01"
        encoded_ref = b"refs/heads/" + name
        with open(os.path.join(os.fsencode(self._repo.path), encoded_ref), "w") as f:
            f.write("00" * 20)

        expected_refs = set(_TEST_REFS.keys())
        expected_refs.add(encoded_ref)

        self.assertEqual(expected_refs, set(self._repo.refs.allkeys()))
        self.assertEqual(
            {r[len(b"refs/") :] for r in expected_refs if r.startswith(b"refs/")},
            set(self._repo.refs.subkeys(b"refs/")),
        )
        expected_refs.remove(b"refs/heads/loop")
        expected_refs.add(b"HEAD")
        self.assertEqual(expected_refs, set(self._repo.get_refs().keys()))

    def test_write_unchanged_ref_optimization(self):
        # Test that writing unchanged ref avoids fsync but still checks locks
        ref_name = b"refs/heads/unchanged"
        ref_value = b"a" * 40

        # Set initial ref value
        self._refs[ref_name] = ref_value

        # Test 1: Writing same value should succeed without changes
        result = self._refs.set_if_equals(ref_name, ref_value, ref_value)
        self.assertTrue(result)

        # Test 2: Writing same value with wrong old_ref should fail
        wrong_old = b"b" * 40
        result = self._refs.set_if_equals(ref_name, wrong_old, ref_value)
        self.assertFalse(result)

        # Test 3: Writing different value should update normally
        new_value = b"c" * 40
        result = self._refs.set_if_equals(ref_name, ref_value, new_value)
        self.assertTrue(result)
        self.assertEqual(new_value, self._refs[ref_name])

    def test_write_unchanged_ref_with_lock(self):
        # Test that file locking is still detected when ref unchanged

        from dulwich.file import FileLocked

        ref_name = b"refs/heads/locktest"
        ref_value = b"d" * 40

        # Set initial ref value
        self._refs[ref_name] = ref_value

        # Get the actual file path
        ref_file = os.path.join(os.fsencode(self._refs.path), ref_name)
        lock_file = ref_file + b".lock"

        # Create lock file to simulate another process holding lock
        with open(lock_file, "wb") as f:
            f.write(b"locked by another process")

            # Try to write same value - should raise FileLocked
            with self.assertRaises(FileLocked):
                self._refs[ref_name] = ref_value

        # Clean up lock file
        if os.path.exists(lock_file):
            os.unlink(lock_file)

        # Now it should work
        self._refs[ref_name] = ref_value


class IsPerWorktreeRefsTests(TestCase):
    def test(self) -> None:
        cases = [
            (b"HEAD", True),
            (b"refs/bisect/good", True),
            (b"refs/worktree/foo", True),
            (b"refs/rewritten/onto", True),
            (b"refs/stash", False),
            (b"refs/heads/main", False),
            (b"refs/tags/v1.0", False),
            (b"refs/remotes/origin/main", False),
            (b"refs/custom/foo", False),
            (b"refs/replace/aaaaaa", False),
        ]
        for ref, expected in cases:
            with self.subTest(ref=ref, expected=expected):
                self.assertEqual(is_per_worktree_ref(ref), expected)


class DiskRefsContainerWorktreeRefsTest(TestCase):
    def setUp(self) -> None:
        # Create temporary directories
        temp_dir = tempfile.mkdtemp()
        test_dir = os.path.join(temp_dir, "main")
        os.makedirs(test_dir)

        repo = Repo.init(test_dir, default_branch=b"main")
        main_worktree = repo.get_worktree()
        with open(os.path.join(test_dir, "test.txt"), "wb") as f:
            f.write(b"test content")
        main_worktree.stage(["test.txt"])
        self.first_commit = main_worktree.commit(message=b"Initial commit")

        worktree_dir = os.path.join(temp_dir, "worktree")
        wt_repo = add_worktree(repo, worktree_dir, branch="wt-main")
        linked_worktree = wt_repo.get_worktree()
        with open(os.path.join(test_dir, "test2.txt"), "wb") as f:
            f.write(b"test content")
        linked_worktree.stage(["test2.txt"])
        self.second_commit = linked_worktree.commit(message=b"second commit")

        self.refs = repo.refs
        self.wt_refs = wt_repo.refs

    def test_refpath(self) -> None:
        main_path = self.refs.path
        common = self.wt_refs.path
        wt_path = self.wt_refs.worktree_path

        cases = [
            (self.refs, b"HEAD", main_path),
            (self.refs, b"refs/heads/main", main_path),
            (self.refs, b"refs/heads/wt-main", main_path),
            (self.refs, b"refs/worktree/foo", main_path),
            (self.refs, b"refs/bisect/good", main_path),
            (self.wt_refs, b"HEAD", wt_path),
            (self.wt_refs, b"refs/heads/main", common),
            (self.wt_refs, b"refs/heads/wt-main", common),
            (self.wt_refs, b"refs/worktree/foo", wt_path),
            (self.wt_refs, b"refs/bisect/good", wt_path),
        ]

        for refs, refname, git_dir in cases:
            with self.subTest(refs=refs, refname=refname, git_dir=git_dir):
                refpath = refs.refpath(refname)
                expected_path = os.path.join(
                    git_dir, refname.replace(b"/", os.fsencode(os.sep))
                )
                self.assertEqual(refpath, expected_path)

    def test_shared_ref(self) -> None:
        self.assertEqual(self.refs[b"refs/heads/main"], self.first_commit)
        self.assertEqual(self.refs[b"refs/heads/wt-main"], self.second_commit)
        self.assertEqual(self.wt_refs[b"refs/heads/main"], self.first_commit)
        self.assertEqual(self.wt_refs[b"refs/heads/wt-main"], self.second_commit)

        expected = {b"HEAD", b"refs/heads/main", b"refs/heads/wt-main"}
        self.assertEqual(expected, self.refs.keys())
        self.assertEqual(expected, self.wt_refs.keys())

        self.assertEqual({b"main", b"wt-main"}, set(self.refs.keys(b"refs/heads/")))
        self.assertEqual({b"main", b"wt-main"}, set(self.wt_refs.keys(b"refs/heads/")))

        ref_path = os.path.join(self.refs.path, b"refs", b"heads", b"main")
        self.assertTrue(os.path.exists(ref_path))

        ref_path = os.path.join(self.wt_refs.worktree_path, b"refs", b"heads", b"main")
        self.assertFalse(os.path.exists(ref_path))

    def test_per_worktree_ref(self) -> None:
        path = self.refs.path
        wt_path = self.wt_refs.worktree_path

        self.assertEqual(self.refs[b"HEAD"], self.first_commit)
        self.assertEqual(self.wt_refs[b"HEAD"], self.second_commit)

        self.refs[b"refs/bisect/good"] = self.first_commit
        self.wt_refs[b"refs/bisect/good"] = self.second_commit

        self.refs[b"refs/bisect/start"] = self.first_commit
        self.wt_refs[b"refs/bisect/bad"] = self.second_commit

        self.assertEqual(self.refs[b"refs/bisect/good"], self.first_commit)
        self.assertEqual(self.wt_refs[b"refs/bisect/good"], self.second_commit)

        self.assertTrue(os.path.exists(os.path.join(path, b"refs", b"bisect", b"good")))
        self.assertTrue(
            os.path.exists(os.path.join(wt_path, b"refs", b"bisect", b"good"))
        )

        self.assertEqual(self.refs[b"refs/bisect/start"], self.first_commit)
        with self.assertRaises(KeyError):
            self.wt_refs[b"refs/bisect/start"]
        self.assertTrue(
            os.path.exists(os.path.join(path, b"refs", b"bisect", b"start"))
        )
        self.assertFalse(
            os.path.exists(os.path.join(wt_path, b"refs", b"bisect", b"start"))
        )

        with self.assertRaises(KeyError):
            self.refs[b"refs/bisect/bad"]
        self.assertEqual(self.wt_refs[b"refs/bisect/bad"], self.second_commit)
        self.assertFalse(os.path.exists(os.path.join(path, b"refs", b"bisect", b"bad")))
        self.assertTrue(
            os.path.exists(os.path.join(wt_path, b"refs", b"bisect", b"bad"))
        )

        expected_refs = {
            b"HEAD",
            b"refs/heads/main",
            b"refs/heads/wt-main",
            b"refs/bisect/good",
            b"refs/bisect/start",
        }
        self.assertEqual(self.refs.keys(), expected_refs)
        self.assertEqual({b"good", b"start"}, self.refs.keys(b"refs/bisect/"))

        expected_wt_refs = {
            b"HEAD",
            b"refs/heads/main",
            b"refs/heads/wt-main",
            b"refs/bisect/good",
            b"refs/bisect/bad",
        }
        self.assertEqual(self.wt_refs.keys(), expected_wt_refs)
        self.assertEqual({b"good", b"bad"}, self.wt_refs.keys(b"refs/bisect/"))

    def test_delete_per_worktree_ref(self) -> None:
        self.refs[b"refs/worktree/foo"] = self.first_commit
        self.wt_refs[b"refs/worktree/foo"] = self.second_commit

        del self.wt_refs[b"refs/worktree/foo"]
        with self.assertRaises(KeyError):
            self.wt_refs[b"refs/worktree/foo"]

        del self.refs[b"refs/worktree/foo"]
        with self.assertRaises(KeyError):
            self.refs[b"refs/worktree/foo"]

    def test_delete_shared_ref(self) -> None:
        self.refs[b"refs/heads/branch"] = self.first_commit

        del self.wt_refs[b"refs/heads/branch"]

        with self.assertRaises(KeyError):
            self.wt_refs[b"refs/heads/branch"]
        with self.assertRaises(KeyError):
            self.refs[b"refs/heads/branch"]

    def test_contains_shared_ref(self):
        self.assertIn(b"refs/heads/main", self.refs)
        self.assertIn(b"refs/heads/main", self.wt_refs)
        self.assertIn(b"refs/heads/wt-main", self.refs)
        self.assertIn(b"refs/heads/wt-main", self.wt_refs)

    def test_contains_per_worktree_ref(self):
        self.refs[b"refs/worktree/foo"] = self.first_commit
        self.wt_refs[b"refs/worktree/bar"] = self.second_commit

        self.assertIn(b"refs/worktree/foo", self.refs)
        self.assertNotIn(b"refs/worktree/bar", self.refs)
        self.assertNotIn(b"refs/worktree/foo", self.wt_refs)
        self.assertIn(b"refs/worktree/bar", self.wt_refs)


_TEST_REFS_SERIALIZED = (
    b"42d06bd4b77fed026b154d16493e5deab78f02ec\t"
    b"refs/heads/40-char-ref-aaaaaaaaaaaaaaaaaa\n"
    b"42d06bd4b77fed026b154d16493e5deab78f02ec\trefs/heads/master\n"
    b"42d06bd4b77fed026b154d16493e5deab78f02ec\trefs/heads/packed\n"
    b"df6800012397fb85c56e7418dd4eb9405dee075c\trefs/tags/refs-0.1\n"
    b"3ec9c43c84ff242e3ef4a9fc5bc111fd780a76a8\trefs/tags/refs-0.2\n"
)


class DiskRefsContainerPathlibTests(TestCase):
    def test_pathlib_init(self) -> None:
        from pathlib import Path

        from dulwich.refs import DiskRefsContainer

        # Create a temporary directory
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(os.rmdir, temp_dir)

        # Test with pathlib.Path
        path_obj = Path(temp_dir)
        refs = DiskRefsContainer(path_obj)
        self.assertEqual(refs.path, temp_dir.encode())

        # Test refpath with pathlib initialized container
        ref_path = refs.refpath(b"HEAD")
        self.assertTrue(isinstance(ref_path, bytes))
        self.assertEqual(ref_path, os.path.join(temp_dir.encode(), b"HEAD"))

    def test_pathlib_worktree_path(self) -> None:
        from pathlib import Path

        from dulwich.refs import DiskRefsContainer

        # Create temporary directories
        temp_dir = tempfile.mkdtemp()
        worktree_dir = tempfile.mkdtemp()
        self.addCleanup(os.rmdir, temp_dir)
        self.addCleanup(os.rmdir, worktree_dir)

        # Test with pathlib.Path for both paths
        path_obj = Path(temp_dir)
        worktree_obj = Path(worktree_dir)
        refs = DiskRefsContainer(path_obj, worktree_path=worktree_obj)
        self.assertEqual(refs.path, temp_dir.encode())
        self.assertEqual(refs.worktree_path, worktree_dir.encode())

        # Test refpath returns worktree path for HEAD
        ref_path = refs.refpath(b"HEAD")
        self.assertEqual(ref_path, os.path.join(worktree_dir.encode(), b"HEAD"))


class InfoRefsContainerTests(TestCase):
    def test_invalid_refname(self) -> None:
        text = _TEST_REFS_SERIALIZED + b"00" * 20 + b"\trefs/stash\n"
        refs = InfoRefsContainer(BytesIO(text))
        expected_refs = dict(_TEST_REFS)
        del expected_refs[b"HEAD"]
        expected_refs[b"refs/stash"] = b"00" * 20
        del expected_refs[b"refs/heads/loop"]
        self.assertEqual(expected_refs, refs.as_dict())

    def test_keys(self) -> None:
        refs = InfoRefsContainer(BytesIO(_TEST_REFS_SERIALIZED))
        actual_keys = set(refs.keys())
        self.assertEqual(set(refs.allkeys()), actual_keys)
        expected_refs = dict(_TEST_REFS)
        del expected_refs[b"HEAD"]
        del expected_refs[b"refs/heads/loop"]
        self.assertEqual(set(expected_refs.keys()), actual_keys)

        actual_keys = refs.keys(b"refs/heads")
        actual_keys.discard(b"loop")
        self.assertEqual(
            [b"40-char-ref-aaaaaaaaaaaaaaaaaa", b"master", b"packed"],
            sorted(actual_keys),
        )
        self.assertEqual([b"refs-0.1", b"refs-0.2"], sorted(refs.keys(b"refs/tags")))

    def test_as_dict(self) -> None:
        refs = InfoRefsContainer(BytesIO(_TEST_REFS_SERIALIZED))
        # refs/heads/loop does not show up even if it exists
        expected_refs = dict(_TEST_REFS)
        del expected_refs[b"HEAD"]
        del expected_refs[b"refs/heads/loop"]
        self.assertEqual(expected_refs, refs.as_dict())

    def test_contains(self) -> None:
        refs = InfoRefsContainer(BytesIO(_TEST_REFS_SERIALIZED))
        self.assertIn(b"refs/heads/master", refs)
        self.assertNotIn(b"refs/heads/bar", refs)

    def test_get_peeled(self) -> None:
        refs = InfoRefsContainer(BytesIO(_TEST_REFS_SERIALIZED))
        # refs/heads/loop does not show up even if it exists
        self.assertEqual(
            _TEST_REFS[b"refs/heads/master"],
            refs.get_peeled(b"refs/heads/master"),
        )


class ParseSymrefValueTests(TestCase):
    def test_valid(self) -> None:
        self.assertEqual(b"refs/heads/foo", parse_symref_value(b"ref: refs/heads/foo"))

    def test_invalid(self) -> None:
        self.assertRaises(ValueError, parse_symref_value, b"foobar")


class ParseRemoteRefTests(TestCase):
    def test_valid(self) -> None:
        # Test simple case
        remote, branch = parse_remote_ref(b"refs/remotes/origin/main")
        self.assertEqual(b"origin", remote)
        self.assertEqual(b"main", branch)

        # Test with branch containing slashes
        remote, branch = parse_remote_ref(b"refs/remotes/upstream/feature/new-ui")
        self.assertEqual(b"upstream", remote)
        self.assertEqual(b"feature/new-ui", branch)

    def test_invalid_not_remote_ref(self) -> None:
        # Not a remote ref
        with self.assertRaises(ValueError) as cm:
            parse_remote_ref(b"refs/heads/main")
        self.assertIn("Not a remote ref", str(cm.exception))

    def test_invalid_format(self) -> None:
        # Missing branch name
        with self.assertRaises(ValueError) as cm:
            parse_remote_ref(b"refs/remotes/origin")
        self.assertIn("Invalid remote ref format", str(cm.exception))

        # Just the prefix
        with self.assertRaises(ValueError) as cm:
            parse_remote_ref(b"refs/remotes/")
        self.assertIn("Invalid remote ref format", str(cm.exception))


class StripPeeledRefsTests(TestCase):
    all_refs: ClassVar[dict[bytes, bytes]] = {
        b"refs/heads/master": b"8843d7f92416211de9ebb963ff4ce28125932878",
        b"refs/heads/testing": b"186a005b134d8639a58b6731c7c1ea821a6eedba",
        b"refs/tags/1.0.0": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
        b"refs/tags/1.0.0^{}": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
        b"refs/tags/2.0.0": b"0749936d0956c661ac8f8d3483774509c165f89e",
        b"refs/tags/2.0.0^{}": b"0749936d0956c661ac8f8d3483774509c165f89e",
    }
    non_peeled_refs: ClassVar[dict[bytes, bytes]] = {
        b"refs/heads/master": b"8843d7f92416211de9ebb963ff4ce28125932878",
        b"refs/heads/testing": b"186a005b134d8639a58b6731c7c1ea821a6eedba",
        b"refs/tags/1.0.0": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
        b"refs/tags/2.0.0": b"0749936d0956c661ac8f8d3483774509c165f89e",
    }

    def test_strip_peeled_refs(self) -> None:
        # Simple check of two dicts
        self.assertEqual(strip_peeled_refs(self.all_refs), self.non_peeled_refs)

    def test_split_peeled_refs(self) -> None:
        (regular, peeled) = split_peeled_refs(self.all_refs)
        self.assertEqual(regular, self.non_peeled_refs)
        self.assertEqual(
            peeled,
            {
                b"refs/tags/2.0.0": b"0749936d0956c661ac8f8d3483774509c165f89e",
                b"refs/tags/1.0.0": b"a93db4b0360cc635a2b93675010bac8d101f73f0",
            },
        )


class ShortenRefNameTests(TestCase):
    """Tests for shorten_ref_name function."""

    def test_branch_ref(self) -> None:
        """Test shortening branch references."""
        self.assertEqual(b"master", shorten_ref_name(b"refs/heads/master"))
        self.assertEqual(b"develop", shorten_ref_name(b"refs/heads/develop"))
        self.assertEqual(
            b"feature/new-ui", shorten_ref_name(b"refs/heads/feature/new-ui")
        )

    def test_remote_ref(self) -> None:
        """Test shortening remote references."""
        self.assertEqual(b"origin/main", shorten_ref_name(b"refs/remotes/origin/main"))
        self.assertEqual(
            b"upstream/master", shorten_ref_name(b"refs/remotes/upstream/master")
        )
        self.assertEqual(
            b"origin/feature/test",
            shorten_ref_name(b"refs/remotes/origin/feature/test"),
        )

    def test_tag_ref(self) -> None:
        """Test shortening tag references."""
        self.assertEqual(b"v1.0", shorten_ref_name(b"refs/tags/v1.0"))
        self.assertEqual(b"release-2.0", shorten_ref_name(b"refs/tags/release-2.0"))

    def test_special_refs(self) -> None:
        """Test that special refs are not shortened."""
        self.assertEqual(b"HEAD", shorten_ref_name(b"HEAD"))
        self.assertEqual(b"FETCH_HEAD", shorten_ref_name(b"FETCH_HEAD"))
        self.assertEqual(b"ORIG_HEAD", shorten_ref_name(b"ORIG_HEAD"))

    def test_other_refs(self) -> None:
        """Test refs that don't match standard prefixes."""
        # Refs that don't match any standard prefix are returned as-is
        self.assertEqual(b"refs/stash", shorten_ref_name(b"refs/stash"))
        self.assertEqual(b"refs/bisect/good", shorten_ref_name(b"refs/bisect/good"))


class RefUtilityFunctionsTests(TestCase):
    """Tests for the new ref utility functions."""

    def test_local_branch_name(self) -> None:
        """Test local_branch_name function."""
        from dulwich.refs import local_branch_name

        # Test adding prefix to branch name
        self.assertEqual(b"refs/heads/master", local_branch_name(b"master"))
        self.assertEqual(b"refs/heads/develop", local_branch_name(b"develop"))
        self.assertEqual(
            b"refs/heads/feature/new-ui", local_branch_name(b"feature/new-ui")
        )

        # Test idempotency - already has prefix
        self.assertEqual(b"refs/heads/master", local_branch_name(b"refs/heads/master"))

    def test_local_tag_name(self) -> None:
        """Test local_tag_name function."""
        from dulwich.refs import local_tag_name

        # Test adding prefix to tag name
        self.assertEqual(b"refs/tags/v1.0", local_tag_name(b"v1.0"))
        self.assertEqual(b"refs/tags/release-2.0", local_tag_name(b"release-2.0"))

        # Test idempotency - already has prefix
        self.assertEqual(b"refs/tags/v1.0", local_tag_name(b"refs/tags/v1.0"))

    def test_extract_branch_name(self) -> None:
        """Test extract_branch_name function."""
        from dulwich.refs import extract_branch_name

        # Test extracting branch name from full ref
        self.assertEqual(b"master", extract_branch_name(b"refs/heads/master"))
        self.assertEqual(b"develop", extract_branch_name(b"refs/heads/develop"))
        self.assertEqual(
            b"feature/new-ui", extract_branch_name(b"refs/heads/feature/new-ui")
        )

        # Test error on invalid ref
        with self.assertRaises(ValueError) as cm:
            extract_branch_name(b"refs/tags/v1.0")
        self.assertIn("Not a local branch ref", str(cm.exception))

        with self.assertRaises(ValueError):
            extract_branch_name(b"master")

    def test_extract_tag_name(self) -> None:
        """Test extract_tag_name function."""
        from dulwich.refs import extract_tag_name

        # Test extracting tag name from full ref
        self.assertEqual(b"v1.0", extract_tag_name(b"refs/tags/v1.0"))
        self.assertEqual(b"release-2.0", extract_tag_name(b"refs/tags/release-2.0"))

        # Test error on invalid ref
        with self.assertRaises(ValueError) as cm:
            extract_tag_name(b"refs/heads/master")
        self.assertIn("Not a local tag ref", str(cm.exception))

        with self.assertRaises(ValueError):
            extract_tag_name(b"v1.0")



# bundle.py -- Bundle format support
# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Bundle format support."""

from collections.abc import Iterator, Sequence
from typing import (
    TYPE_CHECKING,
    BinaryIO,
    Callable,
    Optional,
    Protocol,
    cast,
    runtime_checkable,
)

from .pack import PackData, UnpackedObject, write_pack_data


@runtime_checkable
class PackDataLike(Protocol):
    """Protocol for objects that behave like PackData."""

    def __len__(self) -> int:
        """Return the number of objects in the pack."""
        ...

    def iter_unpacked(self) -> Iterator[UnpackedObject]:
        """Iterate over unpacked objects in the pack."""
        ...


if TYPE_CHECKING:
    from .object_store import BaseObjectStore
    from .repo import BaseRepo


class Bundle:
    """Git bundle object representation."""

    version: Optional[int]

    capabilities: dict[str, Optional[str]]
    prerequisites: list[tuple[bytes, bytes]]
    references: dict[bytes, bytes]
    pack_data: Optional[PackDataLike]

    def __repr__(self) -> str:
        """Return string representation of Bundle."""
        return (
            f"<{type(self).__name__}(version={self.version}, "
            f"capabilities={self.capabilities}, "
            f"prerequisites={self.prerequisites}, "
            f"references={self.references})>"
        )

    def __eq__(self, other: object) -> bool:
        """Check equality with another Bundle."""
        if not isinstance(other, type(self)):
            return False
        if self.version != other.version:
            return False
        if self.capabilities != other.capabilities:
            return False
        if self.prerequisites != other.prerequisites:
            return False
        if self.references != other.references:
            return False
        if self.pack_data != other.pack_data:
            return False
        return True

    def store_objects(
        self,
        object_store: "BaseObjectStore",
        progress: Optional[Callable[[str], None]] = None,
    ) -> None:
        """Store all objects from this bundle into an object store.

        Args:
            object_store: The object store to add objects to
            progress: Optional progress callback function
        """
        from .objects import ShaFile

        if self.pack_data is None:
            raise ValueError("pack_data is not loaded")
        count = 0
        for unpacked in self.pack_data.iter_unpacked():
            # Convert the unpacked object to a proper git object
            if unpacked.decomp_chunks and unpacked.obj_type_num is not None:
                git_obj = ShaFile.from_raw_chunks(
                    unpacked.obj_type_num, unpacked.decomp_chunks
                )
                object_store.add_object(git_obj)
                count += 1

                if progress and count % 100 == 0:
                    progress(f"Stored {count} objects")

        if progress:
            progress(f"Stored {count} objects total")


def _read_bundle(f: BinaryIO, version: int) -> Bundle:
    capabilities = {}
    prerequisites = []
    references = {}
    line = f.readline()
    if version >= 3:
        while line.startswith(b"@"):
            line = line[1:].rstrip(b"\n")
            try:
                key, value_bytes = line.split(b"=", 1)
                value = value_bytes.decode("utf-8")
            except ValueError:
                key = line
                value = None
            capabilities[key.decode("utf-8")] = value
            line = f.readline()
    while line.startswith(b"-"):
        (obj_id, comment) = line[1:].rstrip(b"\n").split(b" ", 1)
        prerequisites.append((obj_id, comment))
        line = f.readline()
    while line != b"\n":
        (obj_id, ref) = line.rstrip(b"\n").split(b" ", 1)
        references[ref] = obj_id
        line = f.readline()
    # Extract pack data to separate stream since PackData expects
    # the file to start with PACK header at position 0
    pack_bytes = f.read()
    if not pack_bytes:
        raise ValueError("Bundle file contains no pack data")

    from io import BytesIO

    pack_file = BytesIO(pack_bytes)
    pack_data = PackData.from_file(pack_file)
    ret = Bundle()
    ret.references = references
    ret.capabilities = capabilities
    ret.prerequisites = prerequisites
    ret.pack_data = pack_data
    ret.version = version
    return ret


def read_bundle(f: BinaryIO) -> Bundle:
    """Read a bundle file.

    Args:
        f: A seekable binary file-like object. The file must remain open
           for the lifetime of the returned Bundle object.
    """
    if not hasattr(f, "seek"):
        raise ValueError("Bundle file must be seekable")

    firstline = f.readline()
    if firstline == b"# v2 git bundle\n":
        return _read_bundle(f, 2)
    if firstline == b"# v3 git bundle\n":
        return _read_bundle(f, 3)
    raise AssertionError(f"unsupported bundle format header: {firstline!r}")


def write_bundle(f: BinaryIO, bundle: Bundle) -> None:
    """Write a bundle to a file.

    Args:
        f: File-like object to write to
        bundle: Bundle object to write
    """
    version = bundle.version
    if version is None:
        if bundle.capabilities:
            version = 3
        else:
            version = 2
    if version == 2:
        f.write(b"# v2 git bundle\n")
    elif version == 3:
        f.write(b"# v3 git bundle\n")
    else:
        raise AssertionError(f"unknown version {version}")
    if version == 3:
        for key, value in bundle.capabilities.items():
            f.write(b"@" + key.encode("utf-8"))
            if value is not None:
                f.write(b"=" + value.encode("utf-8"))
            f.write(b"\n")
    for obj_id, comment in bundle.prerequisites:
        f.write(b"-" + obj_id + b" " + comment + b"\n")
    for ref, obj_id in bundle.references.items():
        f.write(obj_id + b" " + ref + b"\n")
    f.write(b"\n")
    if bundle.pack_data is None:
        raise ValueError("bundle.pack_data is not loaded")
    write_pack_data(
        cast(Callable[[bytes], None], f.write),
        num_records=len(bundle.pack_data),
        records=bundle.pack_data.iter_unpacked(),
    )


def create_bundle_from_repo(
    repo: "BaseRepo",
    refs: Optional[Sequence[bytes]] = None,
    prerequisites: Optional[Sequence[bytes]] = None,
    version: Optional[int] = None,
    capabilities: Optional[dict[str, Optional[str]]] = None,
    progress: Optional[Callable[[str], None]] = None,
) -> Bundle:
    """Create a bundle from a repository.

    Args:
        repo: Repository object to create bundle from
        refs: List of refs to include (defaults to all refs)
        prerequisites: List of commit SHAs that are prerequisites
        version: Bundle version (2 or 3, auto-detected if None)
        capabilities: Bundle capabilities (for v3 bundles)
        progress: Optional progress reporting function

    Returns:
        Bundle object ready for writing
    """
    if refs is None:
        refs = list(repo.refs.keys())

    if prerequisites is None:
        prerequisites = []

    if capabilities is None:
        capabilities = {}

    # Build the references dictionary for the bundle
    bundle_refs = {}
    want_objects = set()

    for ref in refs:
        if ref in repo.refs:
            ref_value = repo.refs[ref]
            # Handle peeled refs
            try:
                peeled_value = repo.refs.get_peeled(ref)
                if peeled_value is not None and peeled_value != ref_value:
                    bundle_refs[ref] = peeled_value
                else:
                    bundle_refs[ref] = ref_value
            except KeyError:
                bundle_refs[ref] = ref_value
            want_objects.add(bundle_refs[ref])

    # Convert prerequisites to proper format
    bundle_prerequisites = []
    have_objects = set()
    for prereq in prerequisites:
        if isinstance(prereq, str):
            prereq = prereq.encode("utf-8")
        if isinstance(prereq, bytes):
            if len(prereq) == 40:  # SHA1 hex string
                try:
                    # Validate it's actually hex
                    bytes.fromhex(prereq.decode("utf-8"))
                    # Store hex in bundle and for pack generation
                    bundle_prerequisites.append((prereq, b""))
                    have_objects.add(prereq)
                except ValueError:
                    # Not a valid hex string, invalid prerequisite
                    raise ValueError(f"Invalid prerequisite format: {prereq!r}")
            elif len(prereq) == 20:
                # Binary SHA, convert to hex for both bundle and pack generation
                hex_prereq = prereq.hex().encode("ascii")
                bundle_prerequisites.append((hex_prereq, b""))
                have_objects.add(hex_prereq)
            else:
                # Invalid length
                raise ValueError(f"Invalid prerequisite SHA length: {len(prereq)}")
        else:
            # Assume it's already a binary SHA
            hex_prereq = prereq.hex().encode("ascii")
            bundle_prerequisites.append((hex_prereq, b""))
            have_objects.append(hex_prereq)

    # Generate pack data containing all objects needed for the refs
    pack_count, pack_objects = repo.generate_pack_data(
        have=have_objects,
        want=want_objects,
        progress=progress,
    )

    # Store the pack objects directly, we'll write them when saving the bundle
    # For now, create a simple wrapper to hold the data
    class _BundlePackData:
        def __init__(self, count: int, objects: Iterator[UnpackedObject]) -> None:
            self._count = count
            self._objects = list(objects)  # Materialize the iterator

        def __len__(self) -> int:
            return self._count

        def iter_unpacked(self) -> Iterator[UnpackedObject]:
            return iter(self._objects)

    pack_data = _BundlePackData(pack_count, pack_objects)

    # Create bundle object
    bundle = Bundle()
    bundle.version = version
    bundle.capabilities = capabilities
    bundle.prerequisites = bundle_prerequisites
    bundle.references = bundle_refs
    bundle.pack_data = pack_data

    return bundle



# merge_drivers.py -- Merge driver support for dulwich
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Merge driver support for dulwich."""

import os
import subprocess
import tempfile
from typing import Callable, Optional, Protocol

from .config import Config


class MergeDriver(Protocol):
    """Protocol for merge drivers."""

    def merge(
        self,
        ancestor: bytes,
        ours: bytes,
        theirs: bytes,
        path: Optional[str] = None,
        marker_size: int = 7,
    ) -> tuple[bytes, bool]:
        """Perform a three-way merge.

        Args:
            ancestor: Content of the common ancestor version
            ours: Content of our version
            theirs: Content of their version
            path: Optional path of the file being merged
            marker_size: Size of conflict markers (default 7)

        Returns:
            Tuple of (merged content, success flag)
            If success is False, the content may contain conflict markers
        """
        ...


class ProcessMergeDriver:
    """Merge driver that runs an external process."""

    def __init__(self, command: str, name: str = "custom"):
        """Initialize process merge driver.

        Args:
            command: Command to run for merging
            name: Name of the merge driver
        """
        self.command = command
        self.name = name

    def merge(
        self,
        ancestor: bytes,
        ours: bytes,
        theirs: bytes,
        path: Optional[str] = None,
        marker_size: int = 7,
    ) -> tuple[bytes, bool]:
        """Perform merge using external process.

        The command is executed with the following placeholders:
        - %O: path to ancestor version (base)
        - %A: path to our version
        - %B: path to their version
        - %L: conflict marker size
        - %P: original path of the file

        The command should write the merge result to the file at %A.
        Exit code 0 means successful merge, non-zero means conflicts.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            # Write temporary files
            ancestor_path = os.path.join(tmpdir, "ancestor")
            ours_path = os.path.join(tmpdir, "ours")
            theirs_path = os.path.join(tmpdir, "theirs")

            with open(ancestor_path, "wb") as f:
                f.write(ancestor)
            with open(ours_path, "wb") as f:
                f.write(ours)
            with open(theirs_path, "wb") as f:
                f.write(theirs)

            # Prepare command with placeholders
            cmd = self.command
            cmd = cmd.replace("%O", ancestor_path)
            cmd = cmd.replace("%A", ours_path)
            cmd = cmd.replace("%B", theirs_path)
            cmd = cmd.replace("%L", str(marker_size))
            if path:
                cmd = cmd.replace("%P", path)

            # Execute merge command
            try:
                result = subprocess.run(
                    cmd,
                    shell=True,
                    capture_output=True,
                    text=False,
                )

                # Read merged content from ours file
                with open(ours_path, "rb") as f:
                    merged_content = f.read()

                # Exit code 0 means clean merge, non-zero means conflicts
                success = result.returncode == 0

                return merged_content, success

            except subprocess.SubprocessError:
                # If the command fails completely, return original with conflicts
                return ours, False


class MergeDriverRegistry:
    """Registry for merge drivers."""

    def __init__(self, config: Optional[Config] = None):
        """Initialize merge driver registry.

        Args:
            config: Git configuration object
        """
        self._drivers: dict[str, MergeDriver] = {}
        self._factories: dict[str, Callable[[], MergeDriver]] = {}
        self._config = config

        # Register built-in drivers
        self._register_builtin_drivers()

    def _register_builtin_drivers(self) -> None:
        """Register built-in merge drivers."""
        # The "text" driver is the default three-way merge
        # We don't register it here as it's handled by the default merge code

    def register_driver(self, name: str, driver: MergeDriver) -> None:
        """Register a merge driver instance.

        Args:
            name: Name of the merge driver
            driver: Driver instance
        """
        self._drivers[name] = driver

    def register_factory(self, name: str, factory: Callable[[], MergeDriver]) -> None:
        """Register a factory function for creating merge drivers.

        Args:
            name: Name of the merge driver
            factory: Factory function that returns a MergeDriver
        """
        self._factories[name] = factory

    def get_driver(self, name: str) -> Optional[MergeDriver]:
        """Get a merge driver by name.

        Args:
            name: Name of the merge driver

        Returns:
            MergeDriver instance or None if not found
        """
        # First check registered drivers
        if name in self._drivers:
            return self._drivers[name]

        # Then check factories
        if name in self._factories:
            driver = self._factories[name]()
            self._drivers[name] = driver
            return driver

        # Finally check configuration
        if self._config:
            config_driver = self._create_from_config(name)
            if config_driver is not None:
                self._drivers[name] = config_driver
                return config_driver

        return None

    def _create_from_config(self, name: str) -> Optional[MergeDriver]:
        """Create a merge driver from git configuration.

        Args:
            name: Name of the merge driver

        Returns:
            MergeDriver instance or None if not configured
        """
        if not self._config:
            return None

        # Look for merge.<name>.driver configuration
        try:
            command = self._config.get(("merge", name), "driver")
            if command:
                return ProcessMergeDriver(command.decode(), name)
        except KeyError:
            pass

        return None


# Global registry instance
_merge_driver_registry: Optional[MergeDriverRegistry] = None


def get_merge_driver_registry(config: Optional[Config] = None) -> MergeDriverRegistry:
    """Get the global merge driver registry.

    Args:
        config: Git configuration object

    Returns:
        MergeDriverRegistry instance
    """
    global _merge_driver_registry
    if _merge_driver_registry is None:
        _merge_driver_registry = MergeDriverRegistry(config)
    elif config is not None:
        # Update config if provided
        _merge_driver_registry._config = config
    return _merge_driver_registry



# test_bitmap.py -- Compatibility tests for git pack bitmaps.
# Copyright (C) 2025 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Compatibility tests for git pack bitmaps."""

import os
import tempfile

from dulwich.bitmap import (
    BITMAP_OPT_FULL_DAG,
    BITMAP_OPT_HASH_CACHE,
    BITMAP_OPT_LOOKUP_TABLE,
    BitmapEntry,
    EWAHBitmap,
    PackBitmap,
    write_bitmap,
)
from dulwich.pack import Pack
from dulwich.repo import Repo

from .. import TestCase
from .utils import remove_ro, require_git_version, rmtree_ro, run_git_or_fail


class BitmapCompatTests(TestCase):
    """Compatibility tests for reading git-generated bitmaps."""

    def setUp(self):
        super().setUp()
        # Git bitmap support was added in 2.0.0
        require_git_version((2, 0, 0))
        self._tempdir = tempfile.mkdtemp()
        self.addCleanup(rmtree_ro, self._tempdir)

    def _init_repo_with_bitmap(self):
        """Create a repo and generate a bitmap using git."""
        repo_path = os.path.join(self._tempdir, "test-repo")
        os.mkdir(repo_path)

        # Initialize repo
        run_git_or_fail(["init"], cwd=repo_path)

        # Create some commits
        test_file = os.path.join(repo_path, "test.txt")
        for i in range(5):
            with open(test_file, "w") as f:
                f.write(f"Content {i}\n")
            run_git_or_fail(["add", "test.txt"], cwd=repo_path)
            run_git_or_fail(
                ["commit", "-m", f"Commit {i}"],
                cwd=repo_path,
                env={"GIT_AUTHOR_NAME": "Test", "GIT_AUTHOR_EMAIL": "test@example.com"},
            )

        # Enable bitmap writing and repack
        run_git_or_fail(
            ["config", "pack.writeBitmaps", "true"],
            cwd=repo_path,
        )
        run_git_or_fail(["repack", "-a", "-d", "-b"], cwd=repo_path)

        return repo_path

    def test_read_git_generated_bitmap(self):
        """Test that Dulwich can read a bitmap generated by git."""
        repo_path = self._init_repo_with_bitmap()

        # Find the pack file with bitmap
        pack_dir = os.path.join(repo_path, ".git", "objects", "pack")
        bitmap_files = [f for f in os.listdir(pack_dir) if f.endswith(".bitmap")]

        # Get the pack file (basename without extension)
        bitmap_name = bitmap_files[0]
        pack_basename = bitmap_name.replace(".bitmap", "")
        pack_path = os.path.join(pack_dir, pack_basename)

        # Verify bitmap file exists at expected location
        bitmap_path = pack_path + ".bitmap"
        self.assertTrue(
            os.path.exists(bitmap_path), f"Bitmap file not found at {bitmap_path}"
        )

        # Try to load the bitmap using Dulwich
        with Pack(pack_path) as pack:
            bitmap = pack.bitmap

            # Basic checks
            self.assertIsNotNone(bitmap, f"Failed to load bitmap from {pack_path}")
            self.assertIsNotNone(bitmap.pack_checksum, "Bitmap missing pack checksum")

            # Check that we have some type bitmaps
            # At minimum, we should have some commits
            self.assertGreater(
                len(bitmap.commit_bitmap.bits),
                0,
                "Commit bitmap should not be empty",
            )

    def test_git_can_use_dulwich_repo_with_bitmap(self):
        """Test that git can work with a repo that has Dulwich-created objects."""
        repo_path = os.path.join(self._tempdir, "dulwich-repo")

        # Create a repo with Dulwich and add commits to ensure git creates bitmaps
        repo = Repo.init(repo_path, mkdir=True)
        self.addCleanup(repo.close)

        # Create actual commits, not just loose objects - git needs commits for bitmaps
        test_file = os.path.join(repo_path, "test.txt")
        for i in range(5):
            with open(test_file, "w") as f:
                f.write(f"Content {i}\n")
            run_git_or_fail(["add", "test.txt"], cwd=repo_path)
            run_git_or_fail(
                ["commit", "-m", f"Commit {i}"],
                cwd=repo_path,
                env={"GIT_AUTHOR_NAME": "Test", "GIT_AUTHOR_EMAIL": "test@example.com"},
            )

        # Configure git to write bitmaps
        run_git_or_fail(
            ["config", "pack.writeBitmaps", "true"],
            cwd=repo_path,
        )

        # Git should be able to repack with bitmaps
        run_git_or_fail(["repack", "-a", "-d", "-b"], cwd=repo_path)

        # Verify git created a bitmap
        pack_dir = os.path.join(repo_path, ".git", "objects", "pack")
        self.assertTrue(os.path.exists(pack_dir), "Pack directory should exist")

        bitmap_files = [f for f in os.listdir(pack_dir) if f.endswith(".bitmap")]
        self.assertGreater(
            len(bitmap_files), 0, "Git should have created a bitmap file after repack"
        )

    def test_git_can_read_dulwich_bitmap(self):
        """Test that git can read a bitmap file written by Dulwich."""
        repo_path = os.path.join(self._tempdir, "dulwich-bitmap-repo")

        # Create a repo with git and add commits
        run_git_or_fail(["init"], cwd=None, env={"GIT_DIR": repo_path})

        test_file = os.path.join(repo_path, "..", "test.txt")
        os.makedirs(os.path.dirname(test_file), exist_ok=True)

        for i in range(5):
            with open(test_file, "w") as f:
                f.write(f"Content {i}\n")
            run_git_or_fail(
                ["add", test_file],
                cwd=os.path.dirname(repo_path),
                env={
                    "GIT_DIR": repo_path,
                    "GIT_WORK_TREE": os.path.dirname(repo_path),
                },
            )
            run_git_or_fail(
                ["commit", "-m", f"Commit {i}"],
                cwd=os.path.dirname(repo_path),
                env={
                    "GIT_DIR": repo_path,
                    "GIT_WORK_TREE": os.path.dirname(repo_path),
                    "GIT_AUTHOR_NAME": "Test",
                    "GIT_AUTHOR_EMAIL": "test@example.com",
                },
            )

        # Create a pack with git first
        run_git_or_fail(["repack", "-a", "-d"], cwd=None, env={"GIT_DIR": repo_path})

        # Now use Dulwich to write a bitmap for the pack
        pack_dir = os.path.join(repo_path, "objects", "pack")
        pack_files = [f for f in os.listdir(pack_dir) if f.endswith(".pack")]
        self.assertGreater(len(pack_files), 0, "Should have at least one pack file")

        pack_basename = pack_files[0].replace(".pack", "")
        pack_path = os.path.join(pack_dir, pack_basename)

        # Load the pack and create bitmap data, then close before writing
        with Pack(pack_path) as pack:
            # Create a simple bitmap for testing
            # Git requires BITMAP_OPT_FULL_DAG flag
            bitmap = PackBitmap(
                flags=BITMAP_OPT_FULL_DAG
                | BITMAP_OPT_HASH_CACHE
                | BITMAP_OPT_LOOKUP_TABLE
            )
            bitmap.pack_checksum = pack.get_stored_checksum()

            # Add bitmap entries for the first few commits in the pack
            for i, (sha, offset, crc) in enumerate(pack.index.iterentries()):
                if i >= 3:  # Just add 3 entries
                    break

                ewah = EWAHBitmap()
                # Mark this object and a couple others as reachable
                for j in range(i + 1):
                    ewah.add(j)

                entry = BitmapEntry(object_pos=i, xor_offset=0, flags=0, bitmap=ewah)
                bitmap.entries[sha] = entry
                bitmap.entries_list.append((sha, entry))

            # Add name hash cache
            bitmap.name_hash_cache = [0x12345678, 0xABCDEF00, 0xFEDCBA98]

        # Write the bitmap after pack is closed to avoid file locking on Windows
        bitmap_path = pack_path + ".bitmap"
        remove_ro(bitmap_path)
        write_bitmap(bitmap_path, bitmap)

        # Verify git can use the repository with our bitmap
        # This should succeed if git can read our bitmap
        run_git_or_fail(
            ["rev-list", "--count", "--use-bitmap-index", "HEAD"],
            cwd=None,
            env={"GIT_DIR": repo_path},
        )

        # Verify git count-objects works with our bitmap
        run_git_or_fail(["count-objects", "-v"], cwd=None, env={"GIT_DIR": repo_path})

    def test_bitmap_file_format_structure(self):
        """Test that git-generated bitmap has expected structure."""
        repo_path = self._init_repo_with_bitmap()

        # Find bitmap
        pack_dir = os.path.join(repo_path, ".git", "objects", "pack")
        bitmap_files = [f for f in os.listdir(pack_dir) if f.endswith(".bitmap")]

        bitmap_path = os.path.join(pack_dir, bitmap_files[0])

        # Read the raw file to verify header
        with open(bitmap_path, "rb") as f:
            signature = f.read(4)
            self.assertEqual(b"BITM", signature, "Invalid bitmap signature")

            version = int.from_bytes(f.read(2), byteorder="big")
            self.assertGreaterEqual(version, 1, "Bitmap version should be >= 1")

        # Load with Dulwich and verify structure
        bitmap_name = bitmap_files[0]
        pack_basename = bitmap_name.replace(".bitmap", "")
        pack_path = os.path.join(pack_dir, pack_basename)
        with Pack(pack_path) as pack:
            bitmap = pack.bitmap

            self.assertIsNotNone(bitmap)
            self.assertEqual(bitmap.version, version)



# Copyright (C) 2017 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Parsing of gitignore files.

For details for the matching rules, see https://git-scm.com/docs/gitignore

Important: When checking if directories are ignored, include a trailing slash in the path.
For example, use "dir/" instead of "dir" to check if a directory is ignored.
"""

import os.path
import re
from collections.abc import Iterable, Sequence
from contextlib import suppress
from typing import TYPE_CHECKING, BinaryIO, Optional, Union

if TYPE_CHECKING:
    from .repo import Repo

from .config import Config, get_xdg_config_home_path


def _pattern_to_str(pattern: Union["Pattern", bytes, str]) -> str:
    """Convert a pattern to string, handling both Pattern objects and raw patterns."""
    if isinstance(pattern, Pattern):
        pattern_data: Union[bytes, str] = pattern.pattern
    else:
        pattern_data = pattern
    return pattern_data.decode() if isinstance(pattern_data, bytes) else pattern_data


def _check_parent_exclusion(path: str, matching_patterns: Sequence["Pattern"]) -> bool:
    """Check if a parent directory exclusion prevents negation patterns from taking effect.

    Args:
        path: Path to check
        matching_patterns: List of Pattern objects that matched the path

    Returns:
        True if parent exclusion applies (negation should be ineffective), False otherwise
    """
    # Find the last negation pattern
    final_negation = next(
        (p for p in reversed(matching_patterns) if not p.is_exclude), None
    )
    if not final_negation:
        return False

    final_pattern_str = _pattern_to_str(final_negation)

    # Check if any exclusion pattern excludes a parent directory
    return any(
        pattern.is_exclude
        and _pattern_excludes_parent(_pattern_to_str(pattern), path, final_pattern_str)
        for pattern in matching_patterns
    )


def _pattern_excludes_parent(
    pattern_str: str, path: str, final_pattern_str: str
) -> bool:
    """Check if a pattern excludes a parent directory of the given path."""
    # Handle **/middle/** patterns
    if pattern_str.startswith("**/") and pattern_str.endswith("/**"):
        middle = pattern_str[3:-3]
        return f"/{middle}/" in f"/{path}" or path.startswith(f"{middle}/")

    # Handle dir/** patterns
    if pattern_str.endswith("/**") and not pattern_str.startswith("**/"):
        base_dir = pattern_str[:-3]
        if not path.startswith(base_dir + "/"):
            return False

        remaining = path[len(base_dir) + 1 :]

        # Special case: dir/** allows immediate child file negations
        if (
            not path.endswith("/")
            and final_pattern_str.startswith("!")
            and "/" not in remaining
        ):
            neg_pattern = final_pattern_str[1:]
            if neg_pattern == path or ("*" in neg_pattern and "**" not in neg_pattern):
                return False

        # Nested files with ** negation patterns
        if "**" in final_pattern_str and Pattern(final_pattern_str[1:].encode()).match(
            path.encode()
        ):
            return False

        return True

    # Directory patterns (ending with /) can exclude parent directories
    if pattern_str.endswith("/") and "/" in path:
        p = Pattern(pattern_str.encode())
        parts = path.split("/")
        return any(
            p.match(("/".join(parts[:i]) + "/").encode()) for i in range(1, len(parts))
        )

    return False


def _translate_segment(segment: bytes) -> bytes:
    """Translate a single path segment to regex, following Git rules exactly."""
    if segment == b"*":
        return b"[^/]+"

    res = b""
    i, n = 0, len(segment)
    while i < n:
        c = segment[i : i + 1]
        i += 1
        if c == b"*":
            res += b"[^/]*"
        elif c == b"?":
            res += b"[^/]"
        elif c == b"\\":
            if i < n:
                res += re.escape(segment[i : i + 1])
                i += 1
            else:
                res += re.escape(c)
        elif c == b"[":
            j = i
            if j < n and segment[j : j + 1] == b"!":
                j += 1
            if j < n and segment[j : j + 1] == b"]":
                j += 1
            while j < n and segment[j : j + 1] != b"]":
                j += 1
            if j >= n:
                res += b"\\["
            else:
                stuff = segment[i:j].replace(b"\\", b"\\\\")
                i = j + 1
                if stuff.startswith(b"!"):
                    stuff = b"^" + stuff[1:]
                elif stuff.startswith(b"^"):
                    stuff = b"\\" + stuff
                res += b"[" + stuff + b"]"
        else:
            res += re.escape(c)
    return res


def _handle_double_asterisk(segments: Sequence[bytes], i: int) -> tuple[bytes, bool]:
    """Handle ** segment processing, returns (regex_part, skip_next)."""
    # Check if ** is at end
    remaining = segments[i + 1 :]
    if all(s == b"" for s in remaining):
        # ** at end - matches everything
        return b".*", False

    # Check if next segment is also **
    if i + 1 < len(segments) and segments[i + 1] == b"**":
        # Consecutive ** segments
        # Check if this ends with a directory pattern (trailing /)
        remaining_after_next = segments[i + 2 :]
        is_dir_pattern = (
            len(remaining_after_next) == 1 and remaining_after_next[0] == b""
        )

        if is_dir_pattern:
            # Pattern like c/**/**/ - requires at least one intermediate directory
            return b"[^/]+/(?:[^/]+/)*", True
        else:
            # Pattern like c/**/**/d - allows zero intermediate directories
            return b"(?:[^/]+/)*", True
    else:
        # ** in middle - handle differently depending on what follows
        if i == 0:
            # ** at start - any prefix
            return b"(?:.*/)??", False
        else:
            # ** in middle - match zero or more complete directory segments
            return b"(?:[^/]+/)*", False


def _handle_leading_patterns(pat: bytes, res: bytes) -> tuple[bytes, bytes]:
    """Handle leading patterns like ``/**/``, ``**/``, or ``/``."""
    if pat.startswith(b"/**/"):
        # Leading /** is same as **
        return pat[4:], b"(.*/)?"
    elif pat.startswith(b"**/"):
        # Leading **/
        return pat[3:], b"(.*/)?"
    elif pat.startswith(b"/"):
        # Leading / means relative to .gitignore location
        return pat[1:], b""
    else:
        return pat, b""


def translate(pat: bytes) -> bytes:
    """Translate a gitignore pattern to a regular expression following Git rules exactly."""
    res = b"(?ms)"

    # Check for invalid patterns with // - Git treats these as broken patterns
    if b"//" in pat:
        # Pattern with // doesn't match anything in Git
        return b"(?!.*)"  # Negative lookahead - matches nothing

    # Don't normalize consecutive ** patterns - Git treats them specially
    # c/**/**/ requires at least one intermediate directory
    # So we keep the pattern as-is

    # Handle patterns with no slashes (match at any level)
    if b"/" not in pat[:-1]:  # No slash except possibly at end
        res += b"(.*/)?"

    # Handle leading patterns
    pat, prefix_added = _handle_leading_patterns(pat, res)
    if prefix_added:
        res += prefix_added

    # Process the rest of the pattern
    if pat == b"**":
        res += b".*"
    else:
        segments = pat.split(b"/")
        i = 0
        while i < len(segments):
            segment = segments[i]

            # Add slash separator (except for first segment)
            if i > 0 and segments[i - 1] != b"**":
                res += re.escape(b"/")

            if segment == b"**":
                regex_part, skip_next = _handle_double_asterisk(segments, i)
                res += regex_part
                if regex_part == b".*":  # End of pattern
                    break
                if skip_next:
                    i += 1
            else:
                res += _translate_segment(segment)

            i += 1

    # Add optional trailing slash for files
    if not pat.endswith(b"/"):
        res += b"/?"

    return res + b"\\Z"


def read_ignore_patterns(f: BinaryIO) -> Iterable[bytes]:
    """Read a git ignore file.

    Args:
      f: File-like object to read from
    Returns: List of patterns
    """
    for line in f:
        line = line.rstrip(b"\r\n")

        # Ignore blank lines, they're used for readability.
        if not line.strip():
            continue

        if line.startswith(b"#"):
            # Comment
            continue

        # Trailing spaces are ignored unless they are quoted with a backslash.
        while line.endswith(b" ") and not line.endswith(b"\\ "):
            line = line[:-1]
        line = line.replace(b"\\ ", b" ")

        yield line


def match_pattern(path: bytes, pattern: bytes, ignorecase: bool = False) -> bool:
    """Match a gitignore-style pattern against a path.

    Args:
      path: Path to match
      pattern: Pattern to match
      ignorecase: Whether to do case-sensitive matching
    Returns:
      bool indicating whether the pattern matched
    """
    return Pattern(pattern, ignorecase).match(path)


class Pattern:
    """A single ignore pattern."""

    def __init__(self, pattern: bytes, ignorecase: bool = False) -> None:
        """Initialize a Pattern object.

        Args:
            pattern: The gitignore pattern as bytes.
            ignorecase: Whether to perform case-insensitive matching.
        """
        self.pattern = pattern
        self.ignorecase = ignorecase

        # Handle negation
        if pattern.startswith(b"!"):
            self.is_exclude = False
            pattern = pattern[1:]
        else:
            # Handle escaping of ! and # at start only
            if (
                pattern.startswith(b"\\")
                and len(pattern) > 1
                and pattern[1:2] in (b"!", b"#")
            ):
                pattern = pattern[1:]
            self.is_exclude = True

        # Check if this is a directory-only pattern
        self.is_directory_only = pattern.endswith(b"/")

        flags = 0
        if self.ignorecase:
            flags = re.IGNORECASE
        self._re = re.compile(translate(pattern), flags)

    def __bytes__(self) -> bytes:
        """Return the pattern as bytes.

        Returns:
            The original pattern as bytes.
        """
        return self.pattern

    def __str__(self) -> str:
        """Return the pattern as a string.

        Returns:
            The pattern decoded as a string.
        """
        return os.fsdecode(self.pattern)

    def __eq__(self, other: object) -> bool:
        """Check equality with another Pattern object.

        Args:
            other: The object to compare with.

        Returns:
            True if patterns and ignorecase flags are equal, False otherwise.
        """
        return (
            isinstance(other, type(self))
            and self.pattern == other.pattern
            and self.ignorecase == other.ignorecase
        )

    def __repr__(self) -> str:
        """Return a string representation of the Pattern object.

        Returns:
            A string representation for debugging.
        """
        return f"{type(self).__name__}({self.pattern!r}, {self.ignorecase!r})"

    def match(self, path: bytes) -> bool:
        """Try to match a path against this ignore pattern.

        Args:
          path: Path to match (relative to ignore location)
        Returns: boolean
        """
        # For negation directory patterns (e.g., !dir/), only match directories
        if self.is_directory_only and not self.is_exclude and not path.endswith(b"/"):
            return False

        # Check if the regex matches
        if self._re.match(path):
            return True

        # For exclusion directory patterns, also match files under the directory
        if (
            self.is_directory_only
            and self.is_exclude
            and not path.endswith(b"/")
            and b"/" in path
        ):
            return bool(self._re.match(path.rsplit(b"/", 1)[0] + b"/"))

        return False


class IgnoreFilter:
    """Filter to apply gitignore patterns.

    Important: When checking if directories are ignored, include a trailing slash.
    For example, use is_ignored("dir/") instead of is_ignored("dir").
    """

    def __init__(
        self,
        patterns: Iterable[bytes],
        ignorecase: bool = False,
        path: Optional[str] = None,
    ) -> None:
        """Initialize an IgnoreFilter with a set of patterns.

        Args:
            patterns: An iterable of gitignore patterns as bytes.
            ignorecase: Whether to perform case-insensitive matching.
            path: Optional path to the ignore file for debugging purposes.
        """
        self._patterns: list[Pattern] = []
        self._ignorecase = ignorecase
        self._path = path
        for pattern in patterns:
            self.append_pattern(pattern)

    def append_pattern(self, pattern: bytes) -> None:
        """Add a pattern to the set."""
        self._patterns.append(Pattern(pattern, self._ignorecase))

    def find_matching(self, path: Union[bytes, str]) -> Iterable[Pattern]:
        """Yield all matching patterns for path.

        Args:
          path: Path to match
        Returns:
          Iterator over iterators
        """
        if not isinstance(path, bytes):
            path = os.fsencode(path)
        for pattern in self._patterns:
            if pattern.match(path):
                yield pattern

    def is_ignored(self, path: Union[bytes, str]) -> Optional[bool]:
        """Check whether a path is ignored using Git-compliant logic.

        For directories, include a trailing slash.

        Returns: status is None if file is not mentioned, True if it is
            included, False if it is explicitly excluded.
        """
        matching_patterns = list(self.find_matching(path))
        if not matching_patterns:
            return None

        # Basic rule: last matching pattern wins
        last_pattern = matching_patterns[-1]
        result = last_pattern.is_exclude

        # Apply Git's parent directory exclusion rule for negations
        if not result:  # Only applies to inclusions (negations)
            result = self._apply_parent_exclusion_rule(
                path.decode() if isinstance(path, bytes) else path, matching_patterns
            )

        return result

    def _apply_parent_exclusion_rule(
        self, path: str, matching_patterns: list[Pattern]
    ) -> bool:
        """Apply Git's parent directory exclusion rule.

        "It is not possible to re-include a file if a parent directory of that file is excluded."
        """
        return _check_parent_exclusion(path, matching_patterns)

    @classmethod
    def from_path(
        cls, path: Union[str, os.PathLike[str]], ignorecase: bool = False
    ) -> "IgnoreFilter":
        """Create an IgnoreFilter from a file path.

        Args:
            path: Path to the ignore file.
            ignorecase: Whether to perform case-insensitive matching.

        Returns:
            An IgnoreFilter instance with patterns loaded from the file.
        """
        with open(path, "rb") as f:
            return cls(read_ignore_patterns(f), ignorecase, path=str(path))

    def __repr__(self) -> str:
        """Return string representation of IgnoreFilter."""
        path = getattr(self, "_path", None)
        if path is not None:
            return f"{type(self).__name__}.from_path({path!r})"
        else:
            return f"<{type(self).__name__}>"


class IgnoreFilterStack:
    """Check for ignore status in multiple filters."""

    def __init__(self, filters: list[IgnoreFilter]) -> None:
        """Initialize an IgnoreFilterStack with multiple filters.

        Args:
            filters: A list of IgnoreFilter objects to check in order.
        """
        self._filters = filters

    def is_ignored(self, path: str) -> Optional[bool]:
        """Check whether a path is explicitly included or excluded in ignores.

        Args:
          path: Path to check
        Returns:
          None if the file is not mentioned, True if it is included,
          False if it is explicitly excluded.
        """
        for filter in self._filters:
            status = filter.is_ignored(path)
            if status is not None:
                return status
        return None

    def __repr__(self) -> str:
        """Return a string representation of the IgnoreFilterStack.

        Returns:
            A string representation for debugging.
        """
        return f"{type(self).__name__}({self._filters!r})"


def default_user_ignore_filter_path(config: Config) -> str:
    """Return default user ignore filter path.

    Args:
      config: A Config object
    Returns:
      Path to a global ignore file
    """
    try:
        value = config.get((b"core",), b"excludesFile")
        assert isinstance(value, bytes)
        return value.decode(encoding="utf-8")
    except KeyError:
        pass

    return get_xdg_config_home_path("git", "ignore")


class IgnoreFilterManager:
    """Ignore file manager with Git-compliant behavior.

    Important: When checking if directories are ignored, include a trailing slash.
    For example, use is_ignored("dir/") instead of is_ignored("dir").
    """

    def __init__(
        self,
        top_path: str,
        global_filters: list[IgnoreFilter],
        ignorecase: bool,
    ) -> None:
        """Initialize an IgnoreFilterManager.

        Args:
            top_path: The top-level directory path to manage ignores for.
            global_filters: List of global ignore filters to apply.
            ignorecase: Whether to perform case-insensitive matching.
        """
        self._path_filters: dict[str, Optional[IgnoreFilter]] = {}
        self._top_path = top_path
        self._global_filters = global_filters
        self._ignorecase = ignorecase

    def __repr__(self) -> str:
        """Return string representation of IgnoreFilterManager."""
        return f"{type(self).__name__}({self._top_path}, {self._global_filters!r}, {self._ignorecase!r})"

    def _load_path(self, path: str) -> Optional[IgnoreFilter]:
        try:
            return self._path_filters[path]
        except KeyError:
            pass

        p = os.path.join(self._top_path, path, ".gitignore")
        try:
            self._path_filters[path] = IgnoreFilter.from_path(p, self._ignorecase)
        except (FileNotFoundError, NotADirectoryError):
            self._path_filters[path] = None
        except OSError as e:
            # On Windows, opening a path that contains a symlink can fail with
            # errno 22 (Invalid argument) when the symlink points outside the repo
            if e.errno == 22:
                self._path_filters[path] = None
            else:
                raise
        return self._path_filters[path]

    def find_matching(self, path: str) -> Iterable[Pattern]:
        """Find matching patterns for path.

        Args:
          path: Path to check
        Returns:
          Iterator over Pattern instances
        """
        if os.path.isabs(path):
            raise ValueError(f"{path} is an absolute path")
        filters = [(0, f) for f in self._global_filters]
        if os.path.sep != "/":
            path = path.replace(os.path.sep, "/")
        parts = path.split("/")
        matches = []
        for i in range(len(parts) + 1):
            dirname = "/".join(parts[:i])
            for s, f in filters:
                relpath = "/".join(parts[s:i])
                if i < len(parts):
                    # Paths leading up to the final part are all directories,
                    # so need a trailing slash.
                    relpath += "/"
                matches += list(f.find_matching(relpath))
            ignore_filter = self._load_path(dirname)
            if ignore_filter is not None:
                filters.insert(0, (i, ignore_filter))
        return iter(matches)

    def is_ignored(self, path: str) -> Optional[bool]:
        """Check whether a path is explicitly included or excluded in ignores.

        Args:
          path: Path to check. For directories, the path should end with '/'.

        Returns:
          None if the file is not mentioned, True if it is included,
          False if it is explicitly excluded.
        """
        matches = list(self.find_matching(path))
        if not matches:
            return None

        # Standard behavior - last matching pattern wins
        result = matches[-1].is_exclude

        # Apply Git's parent directory exclusion rule for negations
        if not result:  # Only check if we would include due to negation
            result = _check_parent_exclusion(path, matches)

        # Apply special case for issue #1203: directory traversal with ** patterns
        if result and path.endswith("/"):
            result = self._apply_directory_traversal_rule(path, matches)

        return result

    def _apply_directory_traversal_rule(
        self, path: str, matches: list["Pattern"]
    ) -> bool:
        """Apply directory traversal rule for issue #1203.

        If a directory would be ignored by a ** pattern, but there are negation
        patterns for its subdirectories, then the directory itself should not
        be ignored (to allow traversal).
        """
        # Original logic for traversal check
        last_excluding_pattern = None
        for match in matches:
            if match.is_exclude:
                last_excluding_pattern = match

        if last_excluding_pattern and (
            last_excluding_pattern.pattern.endswith(b"**")
            or b"**" in last_excluding_pattern.pattern
        ):
            # Check if subdirectories would be unignored
            test_subdir = path + "test/"
            test_matches = list(self.find_matching(test_subdir))
            if test_matches:
                # Use standard logic for test case - last matching pattern wins
                test_result = test_matches[-1].is_exclude
                if test_result is False:
                    return False

        return True  # Keep original result

    @classmethod
    def from_repo(cls, repo: "Repo") -> "IgnoreFilterManager":
        """Create a IgnoreFilterManager from a repository.

        Args:
          repo: Repository object
        Returns:
          A `IgnoreFilterManager` object
        """
        global_filters = []
        for p in [
            os.path.join(repo.controldir(), "info", "exclude"),
            default_user_ignore_filter_path(repo.get_config_stack()),
        ]:
            with suppress(OSError):
                global_filters.append(IgnoreFilter.from_path(os.path.expanduser(p)))
        config = repo.get_config_stack()
        ignorecase = config.get_boolean((b"core"), (b"ignorecase"), False)
        return cls(repo.path, global_filters, ignorecase)



# test_whitespace.py -- Tests for whitespace error detection
# Copyright (C) 2025 Dulwich contributors
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Tests for whitespace error detection."""

from dulwich.whitespace import (
    DEFAULT_WHITESPACE_ERRORS,
    WhitespaceChecker,
    fix_whitespace_errors,
    parse_whitespace_config,
)

from . import TestCase


class WhitespaceConfigTests(TestCase):
    """Test core.whitespace configuration parsing."""

    def test_parse_default(self) -> None:
        """Test default whitespace configuration."""
        errors, tab_width = parse_whitespace_config(None)
        self.assertEqual(errors, DEFAULT_WHITESPACE_ERRORS)
        self.assertEqual(tab_width, 8)

    def test_parse_empty(self) -> None:
        """Test empty whitespace configuration."""
        errors, tab_width = parse_whitespace_config("")
        self.assertEqual(errors, set())
        self.assertEqual(tab_width, 8)

    def test_parse_single_error(self) -> None:
        """Test single error type."""
        errors, tab_width = parse_whitespace_config("blank-at-eol")
        self.assertEqual(errors, {"blank-at-eol"})
        self.assertEqual(tab_width, 8)

    def test_parse_multiple_errors(self) -> None:
        """Test multiple error types."""
        errors, tab_width = parse_whitespace_config(
            "blank-at-eol,space-before-tab,tab-in-indent"
        )
        self.assertEqual(errors, {"blank-at-eol", "space-before-tab", "tab-in-indent"})
        self.assertEqual(tab_width, 8)

    def test_parse_with_negation(self) -> None:
        """Test negation of default errors."""
        errors, tab_width = parse_whitespace_config("-blank-at-eol")
        # Should have defaults minus blank-at-eol
        expected = DEFAULT_WHITESPACE_ERRORS - {"blank-at-eol"}
        self.assertEqual(errors, expected)
        self.assertEqual(tab_width, 8)

    def test_parse_trailing_space_alias(self) -> None:
        """Test that trailing-space is an alias for blank-at-eol."""
        errors, tab_width = parse_whitespace_config("trailing-space")
        self.assertEqual(errors, {"blank-at-eol"})
        self.assertEqual(tab_width, 8)

    def test_parse_tabwidth(self) -> None:
        """Test tabwidth setting."""
        errors, tab_width = parse_whitespace_config("blank-at-eol,tabwidth=4")
        self.assertEqual(errors, {"blank-at-eol"})
        self.assertEqual(tab_width, 4)

    def test_parse_invalid_tabwidth(self) -> None:
        """Test invalid tabwidth defaults to 8."""
        _errors, tab_width = parse_whitespace_config("tabwidth=invalid")
        self.assertEqual(tab_width, 8)

        _errors, tab_width = parse_whitespace_config("tabwidth=0")
        self.assertEqual(tab_width, 8)


class WhitespaceCheckerTests(TestCase):
    """Test WhitespaceChecker functionality."""

    def test_blank_at_eol(self) -> None:
        """Test detection of trailing whitespace."""
        checker = WhitespaceChecker({"blank-at-eol"})

        # No trailing whitespace
        errors = checker.check_line(b"normal line", 1)
        self.assertEqual(errors, [])

        # Trailing space
        errors = checker.check_line(b"trailing space ", 1)
        self.assertEqual(errors, [("blank-at-eol", 1)])

        # Trailing tab
        errors = checker.check_line(b"trailing tab\t", 1)
        self.assertEqual(errors, [("blank-at-eol", 1)])

        # Multiple trailing whitespace
        errors = checker.check_line(b"multiple  \t ", 1)
        self.assertEqual(errors, [("blank-at-eol", 1)])

    def test_space_before_tab(self) -> None:
        """Test detection of space before tab in indentation."""
        checker = WhitespaceChecker({"space-before-tab"})

        # No space before tab
        errors = checker.check_line(b"\tindented", 1)
        self.assertEqual(errors, [])

        # Space before tab in indentation
        errors = checker.check_line(b" \tindented", 1)
        self.assertEqual(errors, [("space-before-tab", 1)])

        # Space before tab not in indentation (should not trigger)
        errors = checker.check_line(b"code \t comment", 1)
        self.assertEqual(errors, [])

    def test_indent_with_non_tab(self) -> None:
        """Test detection of 8+ spaces at start of line."""
        checker = WhitespaceChecker({"indent-with-non-tab"}, tab_width=8)

        # Less than 8 spaces
        errors = checker.check_line(b"    code", 1)
        self.assertEqual(errors, [])

        # Exactly 8 spaces
        errors = checker.check_line(b"        code", 1)
        self.assertEqual(errors, [("indent-with-non-tab", 1)])

        # More than 8 spaces
        errors = checker.check_line(b"         code", 1)
        self.assertEqual(errors, [("indent-with-non-tab", 1)])

        # Tab after spaces resets count
        errors = checker.check_line(b"    \t    code", 1)
        self.assertEqual(errors, [])

        # Custom tab width
        checker = WhitespaceChecker({"indent-with-non-tab"}, tab_width=4)
        errors = checker.check_line(b"    code", 1)
        self.assertEqual(errors, [("indent-with-non-tab", 1)])

    def test_tab_in_indent(self) -> None:
        """Test detection of tabs in indentation."""
        checker = WhitespaceChecker({"tab-in-indent"})

        # No tabs
        errors = checker.check_line(b"    code", 1)
        self.assertEqual(errors, [])

        # Tab in indentation
        errors = checker.check_line(b"\tcode", 1)
        self.assertEqual(errors, [("tab-in-indent", 1)])

        # Tab after non-whitespace (should not trigger)
        errors = checker.check_line(b"code\tcomment", 1)
        self.assertEqual(errors, [])

    def test_cr_at_eol(self) -> None:
        """Test detection of carriage return at end of line."""
        checker = WhitespaceChecker({"cr-at-eol"})

        # No CR
        errors = checker.check_line(b"normal line", 1)
        self.assertEqual(errors, [])

        # CR at end
        errors = checker.check_line(b"line\r", 1)
        self.assertEqual(errors, [("cr-at-eol", 1)])

    def test_blank_at_eof(self) -> None:
        """Test detection of blank lines at end of file."""
        checker = WhitespaceChecker({"blank-at-eof"})

        # No trailing blank lines
        content = b"line1\nline2\nline3"
        errors = checker.check_content(content)
        self.assertEqual(errors, [])

        # One trailing blank line (normal for files ending with newline)
        content = b"line1\nline2\nline3\n"
        errors = checker.check_content(content)
        self.assertEqual(errors, [])

        # Multiple trailing blank lines
        content = b"line1\nline2\n\n\n"
        errors = checker.check_content(content)
        self.assertEqual(errors, [("blank-at-eof", 5)])

        # Only blank lines
        content = b"\n\n\n"
        errors = checker.check_content(content)
        self.assertEqual(errors, [("blank-at-eof", 4)])

    def test_multiple_errors(self) -> None:
        """Test detection of multiple error types."""
        checker = WhitespaceChecker(
            {"blank-at-eol", "space-before-tab", "tab-in-indent"}
        )

        # Line with multiple errors
        errors = checker.check_line(b" \tcode  ", 1)
        error_types = {e[0] for e in errors}
        self.assertEqual(
            error_types, {"blank-at-eol", "space-before-tab", "tab-in-indent"}
        )

    def test_check_content_crlf(self) -> None:
        """Test content checking with CRLF line endings."""
        checker = WhitespaceChecker({"blank-at-eol", "cr-at-eol"})

        # CRLF line endings
        content = b"line1\r\nline2 \r\nline3\r\n"
        errors = checker.check_content(content)
        # Should detect trailing space on line 2 but not CR (since CRLF is handled)
        self.assertEqual(errors, [("blank-at-eol", 2)])


class WhitespaceFixTests(TestCase):
    """Test whitespace error fixing."""

    def test_fix_blank_at_eol(self) -> None:
        """Test fixing trailing whitespace."""
        content = b"line1  \nline2\t\nline3"
        errors = [("blank-at-eol", 1), ("blank-at-eol", 2)]
        fixed = fix_whitespace_errors(content, errors)
        self.assertEqual(fixed, b"line1\nline2\nline3")

    def test_fix_blank_at_eof(self) -> None:
        """Test fixing blank lines at end of file."""
        content = b"line1\nline2\n\n\n"
        errors = [("blank-at-eof", 4)]
        fixed = fix_whitespace_errors(content, errors)
        self.assertEqual(fixed, b"line1\nline2\n")

    def test_fix_cr_at_eol(self) -> None:
        """Test fixing carriage returns."""
        content = b"line1\r\nline2\r\nline3\r"
        errors = [("cr-at-eol", 1), ("cr-at-eol", 2), ("cr-at-eol", 3)]
        fixed = fix_whitespace_errors(content, errors)
        # Our fix function removes all CRs when cr-at-eol errors are fixed
        self.assertEqual(fixed, b"line1\nline2\nline3")

    def test_fix_specific_types(self) -> None:
        """Test fixing only specific error types."""
        content = b"line1  \nline2\n\n\n"
        errors = [("blank-at-eol", 1), ("blank-at-eof", 4)]

        # Fix only blank-at-eol
        fixed = fix_whitespace_errors(content, errors, fix_types={"blank-at-eol"})
        self.assertEqual(fixed, b"line1\nline2\n\n\n")

        # Fix only blank-at-eof
        fixed = fix_whitespace_errors(content, errors, fix_types={"blank-at-eof"})
        self.assertEqual(fixed, b"line1  \nline2\n")

    def test_fix_no_errors(self) -> None:
        """Test fixing with no errors returns original content."""
        content = b"line1\nline2\nline3"
        fixed = fix_whitespace_errors(content, [])
        self.assertEqual(fixed, content)



#!/usr/bin/python3
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later

import tempfile

from google.cloud import storage  # type: ignore[attr-defined]

from dulwich.cloud.gcs import GcsObjectStore
from dulwich.repo import Repo

client = storage.Client()
bucket = client.get_bucket("mybucket")

gcs_object_store = GcsObjectStore(bucket, "path")
r = Repo.init_bare(tempfile.mkdtemp(), object_store=gcs_object_store)



# lfs.py -- Implementation of the LFS
# Copyright (C) 2020 Jelmer Vernooij
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#

"""Git Large File Storage (LFS) support.

This module provides support for Git LFS, which is a Git extension for
versioning large files. It replaces large files with text pointers inside Git,
while storing the file contents on a remote server.

Key components:
- LFS pointer file parsing and creation
- LFS object storage and retrieval
- HTTP client for LFS server communication
- Integration with dulwich repositories
"""

import hashlib
import json
import logging
import os
import tempfile
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, BinaryIO, Optional, Union
from urllib.parse import urljoin, urlparse
from urllib.request import Request, urlopen

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    import urllib3

    from .config import Config
    from .repo import Repo


@dataclass
class LFSAction:
    """LFS action structure."""

    href: str
    header: Optional[dict[str, str]] = None
    expires_at: Optional[str] = None


@dataclass
class LFSErrorInfo:
    """LFS error structure."""

    code: int
    message: str


@dataclass
class LFSBatchObject:
    """LFS batch object structure."""

    oid: str
    size: int
    authenticated: Optional[bool] = None
    actions: Optional[dict[str, LFSAction]] = None
    error: Optional[LFSErrorInfo] = None


@dataclass
class LFSBatchResponse:
    """LFS batch response structure."""

    transfer: str
    objects: list[LFSBatchObject]
    hash_algo: Optional[str] = None


class LFSStore:
    """Stores objects on disk, indexed by SHA256."""

    def __init__(self, path: str) -> None:
        """Initialize LFSStore."""
        self.path = path

    @classmethod
    def create(cls, lfs_dir: str) -> "LFSStore":
        """Create a new LFS store."""
        if not os.path.isdir(lfs_dir):
            os.mkdir(lfs_dir)
        tmp_dir = os.path.join(lfs_dir, "tmp")
        if not os.path.isdir(tmp_dir):
            os.mkdir(tmp_dir)
        objects_dir = os.path.join(lfs_dir, "objects")
        if not os.path.isdir(objects_dir):
            os.mkdir(objects_dir)
        return cls(lfs_dir)

    @classmethod
    def from_repo(cls, repo: "Repo", create: bool = False) -> "LFSStore":
        """Create LFS store from repository."""
        lfs_dir = os.path.join(repo.controldir(), "lfs")
        if create:
            return cls.create(lfs_dir)
        return cls(lfs_dir)

    @classmethod
    def from_controldir(cls, controldir: str, create: bool = False) -> "LFSStore":
        """Create LFS store from control directory."""
        lfs_dir = os.path.join(controldir, "lfs")
        if create:
            return cls.create(lfs_dir)
        return cls(lfs_dir)

    def _sha_path(self, sha: str) -> str:
        return os.path.join(self.path, "objects", sha[0:2], sha[2:4], sha)

    def open_object(self, sha: str) -> BinaryIO:
        """Open an object by sha."""
        try:
            return open(self._sha_path(sha), "rb")
        except FileNotFoundError as exc:
            raise KeyError(sha) from exc

    def write_object(self, chunks: Iterable[bytes]) -> str:
        """Write an object.

        Returns: object SHA
        """
        # First pass: compute SHA256 and collect data
        sha = hashlib.sha256()
        data_chunks = []
        for chunk in chunks:
            sha.update(chunk)
            data_chunks.append(chunk)

        sha_hex = sha.hexdigest()
        path = self._sha_path(sha_hex)

        # If object already exists, no need to write
        if os.path.exists(path):
            return sha_hex

        # Object doesn't exist, write it
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        tmpdir = os.path.join(self.path, "tmp")
        with tempfile.NamedTemporaryFile(dir=tmpdir, mode="wb", delete=False) as f:
            for chunk in data_chunks:
                f.write(chunk)
            f.flush()
            tmppath = f.name

        # Handle concurrent writes - if file already exists, just remove temp file
        if os.path.exists(path):
            os.remove(tmppath)
        else:
            os.rename(tmppath, path)
        return sha_hex


class LFSPointer:
    """Represents an LFS pointer file."""

    def __init__(self, oid: str, size: int) -> None:
        """Initialize LFSPointer."""
        self.oid = oid
        self.size = size

    @classmethod
    def from_bytes(cls, data: bytes) -> Optional["LFSPointer"]:
        """Parse LFS pointer from bytes.

        Returns None if data is not a valid LFS pointer.
        """
        try:
            text = data.decode("utf-8")
        except UnicodeDecodeError:
            return None

        # LFS pointer files have a specific format
        lines = text.strip().split("\n")
        if len(lines) < 3:
            return None

        # Must start with version
        if not lines[0].startswith("version https://git-lfs.github.com/spec/v1"):
            return None

        oid = None
        size = None

        for line in lines[1:]:
            if line.startswith("oid sha256:"):
                oid = line[11:].strip()
            elif line.startswith("size "):
                try:
                    size = int(line[5:].strip())
                    # Size must be non-negative
                    if size < 0:
                        return None
                except ValueError:
                    return None

        if oid is None or size is None:
            return None

        return cls(oid, size)

    def to_bytes(self) -> bytes:
        """Convert LFS pointer to bytes."""
        return (
            f"version https://git-lfs.github.com/spec/v1\n"
            f"oid sha256:{self.oid}\n"
            f"size {self.size}\n"
        ).encode()

    def is_valid_oid(self) -> bool:
        """Check if the OID is valid SHA256."""
        if len(self.oid) != 64:
            return False
        try:
            int(self.oid, 16)
            return True
        except ValueError:
            return False


class LFSFilterDriver:
    """LFS filter driver implementation."""

    def __init__(
        self, lfs_store: "LFSStore", config: Optional["Config"] = None
    ) -> None:
        """Initialize LFSFilterDriver."""
        self.lfs_store = lfs_store
        self.config = config

    def clean(self, data: bytes) -> bytes:
        """Convert file content to LFS pointer (clean filter)."""
        # Check if data is already an LFS pointer
        pointer = LFSPointer.from_bytes(data)
        if pointer is not None:
            return data

        # Store the file content in LFS
        sha = self.lfs_store.write_object([data])

        # Create and return LFS pointer
        pointer = LFSPointer(sha, len(data))
        return pointer.to_bytes()

    def smudge(self, data: bytes, path: bytes = b"") -> bytes:
        """Convert LFS pointer to file content (smudge filter)."""
        # Try to parse as LFS pointer
        pointer = LFSPointer.from_bytes(data)
        if pointer is None:
            # Not an LFS pointer, return as-is
            return data

        # Validate the pointer
        if not pointer.is_valid_oid():
            return data

        try:
            # Read the actual content from LFS store
            with self.lfs_store.open_object(pointer.oid) as f:
                return f.read()
        except KeyError:
            # Object not found in LFS store, try to download it
            try:
                content = self._download_object(pointer)
                return content
            except LFSError as e:
                # Download failed, fall back to returning pointer
                logger.warning("LFS object download failed for %s: %s", pointer.oid, e)

                # Return pointer as-is when object is missing and download failed
                return data

    def _download_object(self, pointer: LFSPointer) -> bytes:
        """Download an LFS object from the server.

        Args:
            pointer: LFS pointer containing OID and size

        Returns:
            Downloaded content

        Raises:
            LFSError: If download fails for any reason
        """
        if self.config is None:
            raise LFSError("No configuration available for LFS download")

        # Create LFS client and download
        client = LFSClient.from_config(self.config)
        if client is None:
            raise LFSError("No LFS client available from configuration")
        content = client.download(pointer.oid, pointer.size)

        # Store the downloaded content in local LFS store
        stored_oid = self.lfs_store.write_object([content])

        # Verify the stored OID matches what we expected
        if stored_oid != pointer.oid:
            raise LFSError(
                f"Downloaded OID mismatch: expected {pointer.oid}, got {stored_oid}"
            )

        return content

    def cleanup(self) -> None:
        """Clean up any resources held by this filter driver."""
        # LFSFilterDriver doesn't hold any resources that need cleanup

    def reuse(self, config: Optional["Config"], filter_name: str) -> bool:
        """Check if this filter driver should be reused with the given configuration."""
        # LFSFilterDriver is stateless and lightweight, no need to cache
        return False


def _get_lfs_user_agent(config: Optional["Config"]) -> str:
    """Get User-Agent string for LFS requests, respecting git config."""
    try:
        if config:
            # Use configured user agent verbatim if set
            return config.get(b"http", b"useragent").decode()
    except KeyError:
        pass

    # Default LFS user agent (similar to git-lfs format)
    from . import __version__

    version_str = ".".join([str(x) for x in __version__])
    return f"git-lfs/dulwich/{version_str}"


def _is_valid_lfs_url(url: str) -> bool:
    """Check if a URL is valid for LFS.

    Git LFS supports http://, https://, and file:// URLs.

    Args:
        url: URL to validate

    Returns:
        True if URL is a valid LFS URL, False otherwise
    """
    parsed = urlparse(url)

    # Must have a scheme
    if not parsed.scheme:
        return False

    # Only support http, https, and file schemes
    if parsed.scheme not in ("http", "https", "file"):
        return False

    # http/https require a hostname
    if parsed.scheme in ("http", "https"):
        return bool(parsed.netloc)

    # file:// URLs must have a path (netloc is typically empty)
    if parsed.scheme == "file":
        return bool(parsed.path)

    return False


class LFSClient:
    """Base class for LFS client operations."""

    def __init__(self, url: str, config: Optional["Config"] = None) -> None:
        """Initialize LFS client.

        Args:
            url: LFS server URL (http://, https://, or file://)
            config: Optional git config for authentication/proxy settings
        """
        self._base_url = url.rstrip("/") + "/"  # Ensure trailing slash for urljoin
        self.config = config

    @property
    def url(self) -> str:
        """Get the LFS server URL without trailing slash."""
        return self._base_url.rstrip("/")

    def download(self, oid: str, size: int, ref: Optional[str] = None) -> bytes:
        """Download an LFS object.

        Args:
            oid: Object ID (SHA256)
            size: Expected size
            ref: Optional ref name

        Returns:
            Object content
        """
        raise NotImplementedError

    def upload(
        self, oid: str, size: int, content: bytes, ref: Optional[str] = None
    ) -> None:
        """Upload an LFS object.

        Args:
            oid: Object ID (SHA256)
            size: Object size
            content: Object content
            ref: Optional ref name
        """
        raise NotImplementedError

    @classmethod
    def from_config(cls, config: "Config") -> Optional["LFSClient"]:
        """Create LFS client from git config.

        Returns the appropriate subclass (HTTPLFSClient or FileLFSClient)
        based on the URL scheme.
        """
        # Try to get LFS URL from config first
        try:
            url = config.get((b"lfs",), b"url").decode()
        except KeyError:
            pass
        else:
            # Validate explicitly configured URL - raise error if invalid
            if not _is_valid_lfs_url(url):
                raise ValueError(
                    f"Invalid lfs.url in config: {url!r}. "
                    "URL must be an absolute URL with scheme http://, https://, or file://."
                )

            # Return appropriate client based on scheme
            parsed = urlparse(url)
            if parsed.scheme in ("http", "https"):
                return HTTPLFSClient(url, config)
            elif parsed.scheme == "file":
                return FileLFSClient(url, config)
            else:
                # This shouldn't happen if _is_valid_lfs_url works correctly
                raise ValueError(f"Unsupported LFS URL scheme: {parsed.scheme}")

        # Fall back to deriving from remote URL (same as git-lfs)
        try:
            remote_url = config.get((b"remote", b"origin"), b"url").decode()
        except KeyError:
            pass
        else:
            # Convert SSH URLs to HTTPS if needed
            if remote_url.startswith("git@"):
                # Convert git@host:user/repo.git to https://host/user/repo.git
                if ":" in remote_url and "/" in remote_url:
                    host_and_path = remote_url[4:]  # Remove "git@"
                    if ":" in host_and_path:
                        host, path = host_and_path.split(":", 1)
                        remote_url = f"https://{host}/{path}"

            # Ensure URL ends with .git for consistent LFS endpoint
            if not remote_url.endswith(".git"):
                remote_url = f"{remote_url}.git"

            # Standard LFS endpoint is remote_url + "/info/lfs"
            lfs_url = f"{remote_url}/info/lfs"

            # Return None if derived URL is invalid (LFS is optional)
            if not _is_valid_lfs_url(lfs_url):
                return None

            # Derived URLs are always http/https
            return HTTPLFSClient(lfs_url, config)

        return None


class HTTPLFSClient(LFSClient):
    """LFS client for HTTP/HTTPS operations."""

    def __init__(self, url: str, config: Optional["Config"] = None) -> None:
        """Initialize HTTP LFS client.

        Args:
            url: LFS server URL (http:// or https://)
            config: Optional git config for authentication/proxy settings
        """
        super().__init__(url, config)
        self._pool_manager: Optional[urllib3.PoolManager] = None

    def _get_pool_manager(self) -> "urllib3.PoolManager":
        """Get urllib3 pool manager with git config applied."""
        if self._pool_manager is None:
            from dulwich.client import default_urllib3_manager

            self._pool_manager = default_urllib3_manager(self.config)
        return self._pool_manager

    def _make_request(
        self,
        method: str,
        path: str,
        data: Optional[bytes] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> bytes:
        """Make an HTTP request to the LFS server."""
        url = urljoin(self._base_url, path)
        req_headers = {
            "Accept": "application/vnd.git-lfs+json",
            "Content-Type": "application/vnd.git-lfs+json",
            "User-Agent": _get_lfs_user_agent(self.config),
        }
        if headers:
            req_headers.update(headers)

        # Use urllib3 pool manager with git config applied
        pool_manager = self._get_pool_manager()
        response = pool_manager.request(method, url, headers=req_headers, body=data)
        if response.status >= 400:
            raise ValueError(
                f"HTTP {response.status}: {response.data.decode('utf-8', errors='ignore')}"
            )
        return response.data

    def batch(
        self,
        operation: str,
        objects: list[dict[str, Union[str, int]]],
        ref: Optional[str] = None,
    ) -> LFSBatchResponse:
        """Perform batch operation to get transfer URLs.

        Args:
            operation: "download" or "upload"
            objects: List of {"oid": str, "size": int} dicts
            ref: Optional ref name

        Returns:
            Batch response from server
        """
        data: dict[
            str, Union[str, list[str], list[dict[str, Union[str, int]]], dict[str, str]]
        ] = {
            "operation": operation,
            "transfers": ["basic"],
            "objects": objects,
        }
        if ref:
            data["ref"] = {"name": ref}

        response = self._make_request(
            "POST", "objects/batch", json.dumps(data).encode("utf-8")
        )
        if not response:
            raise ValueError("Empty response from LFS server")
        response_data = json.loads(response)
        return self._parse_batch_response(response_data)

    def _parse_batch_response(self, data: Mapping[str, Any]) -> LFSBatchResponse:
        """Parse JSON response into LFSBatchResponse dataclass."""
        objects = []
        for obj_data in data.get("objects", []):
            actions = None
            if "actions" in obj_data:
                actions = {}
                for action_name, action_data in obj_data["actions"].items():
                    actions[action_name] = LFSAction(
                        href=action_data["href"],
                        header=action_data.get("header"),
                        expires_at=action_data.get("expires_at"),
                    )

            error = None
            if "error" in obj_data:
                error = LFSErrorInfo(
                    code=obj_data["error"]["code"], message=obj_data["error"]["message"]
                )

            batch_obj = LFSBatchObject(
                oid=obj_data["oid"],
                size=obj_data["size"],
                authenticated=obj_data.get("authenticated"),
                actions=actions,
                error=error,
            )
            objects.append(batch_obj)

        return LFSBatchResponse(
            transfer=data.get("transfer", "basic"),
            objects=objects,
            hash_algo=data.get("hash_algo"),
        )

    def download(self, oid: str, size: int, ref: Optional[str] = None) -> bytes:
        """Download an LFS object.

        Args:
            oid: Object ID (SHA256)
            size: Expected size
            ref: Optional ref name

        Returns:
            Object content
        """
        # Get download URL via batch API
        batch_resp = self.batch("download", [{"oid": oid, "size": size}], ref)

        if not batch_resp.objects:
            raise LFSError(f"No objects returned for {oid}")

        obj = batch_resp.objects[0]
        if obj.error:
            raise LFSError(f"Server error for {oid}: {obj.error.message}")

        if not obj.actions or "download" not in obj.actions:
            raise LFSError(f"No download actions for {oid}")

        download_action = obj.actions["download"]
        download_url = download_action.href

        # Download the object using urllib3 with git config
        download_headers = {"User-Agent": _get_lfs_user_agent(self.config)}
        if download_action.header:
            download_headers.update(download_action.header)

        pool_manager = self._get_pool_manager()
        response = pool_manager.request("GET", download_url, headers=download_headers)
        content = response.data

        # Verify size
        if len(content) != size:
            raise LFSError(f"Downloaded size {len(content)} != expected {size}")

        # Verify SHA256
        actual_oid = hashlib.sha256(content).hexdigest()
        if actual_oid != oid:
            raise LFSError(f"Downloaded OID {actual_oid} != expected {oid}")

        return content

    def upload(
        self, oid: str, size: int, content: bytes, ref: Optional[str] = None
    ) -> None:
        """Upload an LFS object.

        Args:
            oid: Object ID (SHA256)
            size: Object size
            content: Object content
            ref: Optional ref name
        """
        # Get upload URL via batch API
        batch_resp = self.batch("upload", [{"oid": oid, "size": size}], ref)

        if not batch_resp.objects:
            raise LFSError(f"No objects returned for {oid}")

        obj = batch_resp.objects[0]
        if obj.error:
            raise LFSError(f"Server error for {oid}: {obj.error.message}")

        # If no actions, object already exists
        if not obj.actions:
            return

        if "upload" not in obj.actions:
            raise LFSError(f"No upload action for {oid}")

        upload_action = obj.actions["upload"]
        upload_url = upload_action.href

        # Upload the object
        req = Request(upload_url, data=content, method="PUT")
        if upload_action.header:
            for name, value in upload_action.header.items():
                req.add_header(name, value)

        with urlopen(req) as response:
            if response.status >= 400:
                raise LFSError(f"Upload failed with status {response.status}")

        # Verify if needed
        if obj.actions and "verify" in obj.actions:
            verify_action = obj.actions["verify"]
            verify_data = json.dumps({"oid": oid, "size": size}).encode("utf-8")

            req = Request(verify_action.href, data=verify_data, method="POST")
            req.add_header("Content-Type", "application/vnd.git-lfs+json")
            if verify_action.header:
                for name, value in verify_action.header.items():
                    req.add_header(name, value)

            with urlopen(req) as response:
                if response.status >= 400:
                    raise LFSError(f"Verification failed with status {response.status}")


class FileLFSClient(LFSClient):
    """LFS client for file:// URLs that accesses local filesystem."""

    def __init__(self, url: str, config: Optional["Config"] = None) -> None:
        """Initialize File LFS client.

        Args:
            url: LFS server URL (file://)
            config: Optional git config (unused for file:// URLs)
        """
        super().__init__(url, config)

        # Convert file:// URL to filesystem path
        from urllib.request import url2pathname

        parsed = urlparse(url)
        if parsed.scheme != "file":
            raise ValueError(f"FileLFSClient requires file:// URL, got {url!r}")

        # url2pathname handles the conversion properly across platforms
        path = url2pathname(parsed.path)
        self._local_store = LFSStore(path)

    def download(self, oid: str, size: int, ref: Optional[str] = None) -> bytes:
        """Download an LFS object from local filesystem.

        Args:
            oid: Object ID (SHA256)
            size: Expected size
            ref: Optional ref name (ignored for file:// URLs)

        Returns:
            Object content

        Raises:
            LFSError: If object not found or size mismatch
        """
        try:
            with self._local_store.open_object(oid) as f:
                content = f.read()
        except KeyError as exc:
            raise LFSError(f"Object not found: {oid}") from exc

        # Verify size
        if len(content) != size:
            raise LFSError(f"Size mismatch: expected {size}, got {len(content)}")

        # Verify SHA256
        actual_oid = hashlib.sha256(content).hexdigest()
        if actual_oid != oid:
            raise LFSError(f"OID mismatch: expected {oid}, got {actual_oid}")

        return content

    def upload(
        self, oid: str, size: int, content: bytes, ref: Optional[str] = None
    ) -> None:
        """Upload an LFS object to local filesystem.

        Args:
            oid: Object ID (SHA256)
            size: Object size
            content: Object content
            ref: Optional ref name (ignored for file:// URLs)

        Raises:
            LFSError: If size or OID mismatch
        """
        # Verify size
        if len(content) != size:
            raise LFSError(f"Size mismatch: expected {size}, got {len(content)}")

        # Verify SHA256
        actual_oid = hashlib.sha256(content).hexdigest()
        if actual_oid != oid:
            raise LFSError(f"OID mismatch: expected {oid}, got {actual_oid}")

        # Store the object
        stored_oid = self._local_store.write_object([content])
        if stored_oid != oid:
            raise LFSError(f"Storage OID mismatch: expected {oid}, got {stored_oid}")


class LFSError(Exception):
    """LFS-specific error."""



# __init__.py -- The git module of dulwich
# Copyright (C) 2007 James Westby <jw+debian@jameswestby.net>
# Copyright (C) 2008 Jelmer Vernooij <jelmer@jelmer.uk>
#
# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
# Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
# General Public License as published by the Free Software Foundation; version 2.0
# or (at your option) any later version. You can redistribute it and/or
# modify it under the terms of either of these two licenses.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# You should have received a copy of the licenses; if not, see
# <http://www.gnu.org/licenses/> for a copy of the GNU General Public License
# and <http://www.apache.org/licenses/LICENSE-2.0> for a copy of the Apache
# License, Version 2.0.
#


"""Python implementation of the Git file formats and protocols."""

import sys
from typing import Any, Callable, Optional, TypeVar, Union

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

__version__ = (0, 24, 7)

__all__ = ["__version__", "replace_me"]

P = ParamSpec("P")
R = TypeVar("R")
F = TypeVar("F", bound=Callable[..., Any])

try:
    from dissolve import replace_me as replace_me
except ImportError:
    # if dissolve is not installed, then just provide a basic implementation
    # of its replace_me decorator
    def replace_me(
        since: Optional[Union[tuple[int, ...], str]] = None,
        remove_in: Optional[Union[tuple[int, ...], str]] = None,
    ) -> Callable[[F], F]:
        """Decorator to mark functions as deprecated.

        Args:
            since: Version when the function was deprecated
            remove_in: Version when the function will be removed

        Returns:
            Decorator function
        """

        def decorator(func: Callable[P, R]) -> Callable[P, R]:
            import functools
            import warnings

            m = f"{func.__name__} is deprecated"
            since_str = str(since) if since is not None else None
            remove_in_str = str(remove_in) if remove_in is not None else None

            if since_str is not None and remove_in_str is not None:
                m += f" since {since_str} and will be removed in {remove_in_str}"
            elif since_str is not None:
                m += f" since {since_str}"
            elif remove_in_str is not None:
                m += f" and will be removed in {remove_in_str}"
            else:
                m += " and will be removed in a future version"

            @functools.wraps(func)
            def _wrapped_func(*args: P.args, **kwargs: P.kwargs) -> R:
                warnings.warn(
                    m,
                    DeprecationWarning,
                    stacklevel=2,
                )
                return func(*args, **kwargs)

            return _wrapped_func

        return decorator  # type: ignore[return-value]
